#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import os
import os.path
import sys
import gobject
from config import *
import storage
import traceback
import util

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_USERPROFILE, fmt % args)

class ModuleLoader:
    """Loads all python modules from a directory allows objects
    to be constructed from said modules using specific constructors
    for each object type."""
    
    def __init__ (self, module_path):
        """Construct a ModuleLoader and will load all available
        python modules from @module_path.
        """
        self.module_path = module_path
        self.modules = []
        self.__load_modules ()

    def __load_module (self, module):
        """Load a python module named @module."""
        dprint ("Loading module: %s" % module)
        cmd = "from sources import %s" % module
        try:
            exec (cmd)
        except:
            print >> sys.stderr, "Failed to import module '%s' cmd=\"%s\"" % (module, cmd)
            traceback.print_exc(file=sys.stderr)
            return
        self.modules.append (module)
        
    def __load_modules (self):
        """Load all available modules from self.module_path."""
        dprint ("Loading modules from %s" % self.module_path)
        sys.path.append (self.module_path)
        for file in os.listdir (self.module_path):
            if file[0] == '.':
                continue
            if file[-3:] != ".py":
                continue
            if file == "__init__.py":
                continue
            self.__load_module (file[:-3])

    def __construct_object (self, module, constructor, arg):
        """Construct an object by invoking a function named @constructor,
        with @arg as an argument, in the module called @module.
        """
        dprint ("Constructing object from loaded module using %s.%s" % (module, constructor))
        cmd = ("from sources import %s\n"
               "if %s.__dict__.has_key ('%s'):"
               "    ret = %s.%s (arg)") % (module, module, constructor, module, constructor)
        try:
            exec (cmd)
        except:
            print "Failed to invoke function '%s.%s': %s" % (module, constructor, sys.exc_type)
            traceback.print_exc(file=sys.stderr)
            return None

        try:
            ret = ret
        except NameError:
            return None
        return ret
    
    def construct_objects (self, constructor, arg):
        """Construct and return a list of objects by invoking a function
        named @constructor, with @arg as an argument, in each of the
        python modules in self.module_path which contain a function
        by that name.
        """
        objects = []
        for module in self.modules:
            o = self.__construct_object (module, constructor, arg)
            if o != None:
                objects.append (o)
        return objects

module_loader = None
def get_module_loader ():
    """Return a singleton ModuleLoader object."""
    global module_loader
    if module_loader == None:
        module_loader = ModuleLoader (os.path.join (os.path.dirname (__file__), "sources"))
    return module_loader
            
class ProfileChange (gobject.GObject):
    """Abstract base class for encapsulating profile changes."""
    
    def __init__ (self, source, delegate = None):
        """Construct a ProfileChange object.

        source: the ProfileSource from which the change came.
        """
        gobject.GObject.__init__ (self)
        self.source = source
        self.delegate = delegate

    def get_source (self):
        """Get the ProfileSource from which this change came."""
        return self.source

    def get_id (self):
        """Return and identifier for the configuration item which changed."""
        raise Exception ("Not implemented")

    def get_short_description (self):
        """Return a short description of the configuration change."""
        raise Exception ("Not implemented")

gobject.type_register (ProfileChange)

class SourceDelegate:
    """An abstract base class for helper classes which can be used
    to intercept and modify changes from a given configuration
    source."""
    
    def __init__ (self, delegate_name, source, namespace_section):
        """Construct a SourceDelegate object.

        @source: the ProfileSource whose changes the delegate wishes
        to inspect.
        @namepsace_section: the section of @source's configuration
        namespace that the delegate wishes to inspect.
        """
        self.name = delegate_name
        self.source = source
        self.namespace_section = namespace_section

    def get_name (self):
        """Returns the configuration delegate's name."""
        return self.name

    def handle_change (self, change):
        """Inspect a ProfileChange. Return #True if the change should
        not be passed on any further (i.e. #True == 'handled') and
        return #False if the change should be passed on unmodified.
        """
        raise Exception ("Not implemented")

    def commit_change (self, change, mandatory = False):
        """Commit a change to profile.

        mandatory: whether the change should be committed such
        that it overrides the user's value.
        """
        raise Exception ("Not implemented")

    def start_monitoring (self):
        """Start monitoring for configuration changes."""
        raise Exception ("Not implemented")

    def stop_monitoring (self):
        """Stop monitoring for configuration changes."""
        raise Exception ("Not implemented")

    def sync_changes (self):
        """Save all committed changes to disk."""
        raise Exception ("Not implemented")

    def apply (self):
        """Apply profile to the current user's environment."""
        raise Exception ("Not implemented")
    
class ProfileSource (gobject.GObject):
    """An abstract base class which each configuration source must
    implement."""

    __gsignals__ = {
        "changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (ProfileChange, ))
        }

    def __init__ (self, source_name, delegate_constructor = None):
        """Construct a ProfileSource object."""
        gobject.GObject.__init__ (self)
        self.name = source_name
        
        module_loader = get_module_loader ()

        self.delegates = []
        if delegate_constructor:
            self.delegates = module_loader.construct_objects (delegate_constructor, self)
    
    def get_name (self):
        """Returns the configuration source's name."""
        return self.name

    def get_path_description (self, path):
        """Returns a human readable description for @path.
        @path must have been previously added by the source to the
        #ProfileStorage.
        """
        return path

    def emit_change (self, change):
        """Pass @change to all register delegates for this source and
        emit a 'changed' signal with @change if none of the delegates
        return #True.
        """
        for delegate in self.delegates:
            if not change.get_id ().startswith (delegate.namespace_section):
                continue
            if delegate.handle_change (change):
                dprint ("Delegate '%s' handled change '%s'" % (delegate.get_name (), change.get_id ()))
                return
        self.emit ("changed", change)
        
    def commit_change (self, change, mandatory = False):
        """Commit a change to profile.

        mandatory: whether the change should be committed such
        that it overrides the user's value.
        """
        if change.delegate:
            change.delegate.commit_change (change, mandatory)
            dprint ("Delegate '%s' committed changes '%s'" % (change.delegate.get_name (), change.get_id ()))
            return True
        return False
    
    def start_monitoring (self):
        """Start monitoring for configuration changes."""
        raise Exception ("Not implemented")

    def stop_monitoring (self):
        """Stop monitoring for configuration changes."""
        raise Exception ("Not implemented")

    def sync_changes (self):
        """Save all committed changes to disk."""
        raise Exception ("Not implemented")

    def apply (self):
        """Apply profile to the current user's environment."""
        raise Exception ("Not implemented")

gobject.type_register (ProfileSource)

class UserProfile (gobject.GObject):
    """An encapsulation of the user profile and backend sources."""
    
    __gsignals__ = {
        "changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (ProfileChange, ))
        }

    def __init__ (self, profile_name):
        """Construct a UserProfile

	profile_name: name of the profile to be loaded
        module_path: optional path from which configuration modules
        should be loaded
        """
        gobject.GObject.__init__ (self)

        dprint ("Constructing profile from %s" % profile_name)

        self.profile_name = profile_name

        #
	# Open the user settings packages and try to install them
	#
	self.storage = storage.ProfileStorage (profile_name)
        
        module_loader = get_module_loader ()
        
        self.sources = []
        self.sources = module_loader.construct_objects ("get_source", self.storage)
        dprint ("%d sources loaded:" % len (self.sources))
        for source in self.sources:
            dprint ("  %s" % source.get_name ())
            source.connect ("changed", self.__handle_source_changed)

    def __handle_source_changed (self, source, change):
        self.emit ("changed", change)

    def get_source (self, source_name):
        """Return a source named @source_name."""
        for source in self.sources:
            if source.get_name () == source_name:
                return source
        return None

    def start_monitoring (self):
        """Start monitoring for configuration changes."""
        dprint ("Starting monitoring")
        for s in self.sources:
            for delegate in s.delegates:
                delegate.start_monitoring ()
            s.start_monitoring ()

    def stop_monitoring (self):
        """Stop monitoring for configuration changes."""
        dprint ("Stopping monitoring")
        for s in self.sources:
            for delegate in s.delegates:
                delegate.stop_monitoring ()
            s.stop_monitoring ()

    def sync_changes (self):
        """Save all committed changes to disk."""
        dprint ("Syncing all committed changes to disk")
        for s in self.sources:
            for delegate in s.delegates:
                delegate.sync_changes ()
            s.sync_changes ()
        self.storage.save ()

    def apply (self):
        """Apply profile to the current user's environment."""
        dprint ("Applying profile to user's environment")
        for s in self.sources:
            s.apply ()
            for delegate in s.delegates:
                delegate.apply ()

gobject.type_register (UserProfile)

#
# Unit tests
#
def run_unit_tests ():
    class LocalTestChange (ProfileChange):
        def __init__ (self, source, key, value):
            ProfileChange.__init__ (self, source)
            self.key = key
            self.value = value
        def get_id (self):
            return self.key
        def get_short_description (self):
            return self.key

    class LocalTestDelegate (SourceDelegate):
        def __init__ (self, source):
            SourceDelegate.__init__ (self, "local", source, "/bar")
        def handle_change (self, change):
            if change.get_id () == "/bar/foo1":
                return True
            return False

    class LocalTestSource (ProfileSource):
        def __init__ (self):
            ProfileSource.__init__ (self, "local")

    profile = UserProfile ("foo-storage")
    assert len (profile.sources) > 0

    testsource = None
    for source in profile.sources:
        if source.get_name () == "test":
            testsource = source
            break;
    assert testsource != None
    assert len (testsource.delegates) == 1

    localsource = LocalTestSource ()
    profile.sources.append (localsource)
    def handle_source_changed (source, change, profile):
        profile.emit ("changed", change)
    localsource.connect ("changed", handle_source_changed, profile)
    assert len (profile.sources) > 1

    localdelegate = LocalTestDelegate (localsource)
    localsource.delegates.append (localdelegate)

    global n_signals
    n_signals = 0
    def handle_changed (profile, change):
        global n_signals
        n_signals += 1
    profile.connect ("changed", handle_changed)

    # Only foo2 and foo3 should get through
    localsource.emit_change (LocalTestChange (localsource, "/bar/foo1", "1"))
    localsource.emit_change (LocalTestChange (localsource, "/bar/foo2", "2"))
    localsource.emit_change (LocalTestChange (localsource, "/bar/foo3", "3"))

    # Only bar2 and bar3 should get through
    testsource.emit_change (testsource.get_test_change ("/foo/bar1", "1"))
    testsource.emit_change (testsource.get_test_change ("/foo/bar2", "2"))
    testsource.emit_change (testsource.get_test_change ("/foo/bar3", "3"))

    assert n_signals == 4



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import os
import os.path
import socket
import errno
import signal
import select
import pwd
import commands
import binascii
import struct
import tempfile
import shutil
import gobject
import util
import usermod
import userprofile
from config import *

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_PROTOSESSION, fmt % args)

class ProtoSessionError (Exception):
    pass

class SessionStartError (ProtoSessionError):
    pass

class XauthParseError (ProtoSessionError):
    pass

def safe_kill (pid, sig):
    try:
        os.kill (pid, sig)
    except os.error, (err, str):
        if err != errno.ESRCH:
            raise

class ProtoSession (gobject.GObject):
    __gsignals__ = {
        "finished" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
        }
    
    def __init__ (self, username, profile_file):
        gobject.GObject.__init__ (self)
        assert os.geteuid () == 0
        
        self.username = username
        self.profile_file = profile_file

        self.usr1_pipe_r = 0
        self.usr1_pipe_w = 0
        
        self.xnest_pid = 0
        self.xnest_child_watch = 0
        self.xnest_xauth_file = None
        
        self.session_pid = 0
        self.session_child_watch = 0
        self.session_xauth_file = None
        
        self.admin_tool_timeout = 0
        self.admin_tool_pid = 0
        self.admin_tool_child_watch = 0

        try:
            pw = pwd.getpwnam (self.username)
        except KeyError:
            raise SessionStartError, _("User '%s' does not exist") % self.username

        self.user_pw = pw
        
    def __kill_xnest (self):
        if self.xnest_child_watch:
            gobject.source_remove (self.xnest_child_watch)
            self.xnest_child_watch = 0
        
        if self.xnest_pid:
            safe_kill (self.xnest_pid, signal.SIGTERM)
            self.xnest_pid = 0

    def __kill_session (self):
        if self.session_child_watch:
            gobject.source_remove (self.session_child_watch)
            self.session_child_watch = 0
        
        if self.session_pid:
            safe_kill (self.session_pid, signal.SIGTERM)
            self.session_pid = 0

    def __kill_admin_tool (self):
        if self.admin_tool_timeout:
            gobject.source_remove (self.admin_tool_timeout)
            self.admin_tool_timeout = 0
        
        if self.admin_tool_child_watch:
            gobject.source_remove (self.admin_tool_child_watch)
            self.admin_tool_child_watch = 0
        
        if self.admin_tool_pid:
            safe_kill (self.admin_tool_pid, signal.SIGTERM)
            self.admin_tool_pid = 0

    def __del__ (self):
        if self.xnest_auth_file:
            os.remove (self.xnest_auth_file)
        if self.session_xauth_file:
            os.remove (self.session_xauth_file)
        if self.usr1_pipe_r:
            close (self.usr1_pipe_r)
        if self.usr1_pipe_w:
            close (self.usr1_pipe_w)
        self.force_quit ()

    def __is_display_free (self, display_number):
        # First make sure we get CONNREFUSED connecting
        # to port 6000 + display_number
        refused = False
        sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect (("127.0.0.1", 6000 + display_number))
        except socket.error, (err, str):
            if err == errno.ECONNREFUSED:
                refused = True
        
        sock.close ()
        if not refused:
            return False

        # Now make sure either that the lock file doesn't exist or
        # that the server specified in the lock file isn't running
        lock_file = "/tmp/.X%d-lock" % display_number
        if os.path.exists (lock_file):
            f = file (lock_file, "r")
            try:
                pid = int (f.read ())
            except ValueError:
                return False

            process_exists = True
            try:
                os.kill (pid, 0)
            except os.error, (err, str):
                if err == errno.ESRCH:
                    process_exists = False
            if process_exists:
                return False

        return True

    def __find_free_display (self):
        display_number = 1
        while display_number < 100:
            if self.__is_display_free (display_number):
                return display_number
            display_number += 1
        return -1

    #
    # This USR1 malarky is that if you set the signal handler for
    # SIGUSR1 to SIG_IGN, then the Xserver will send its parent
    # a SIGUSR1 when its ready to start accepting connections.
    #
    def __sigusr1_handler (self, sig, frame):
        self.got_usr1_signal = True
        os.write (self.usr1_pipe_w, "Y")

    def __xnest_child_watch_handler (self, pid, status):
        dprint ("Xnest died")
        
        self.xnest_pid = 0
        self.xnest_child_watch = 0

        self.force_quit ()

        # If we're waiting for USR1, quit main loop
        if self.main_loop:
            self.main_loop.quit ()

        self.emit ("finished");
        
        return False

    def __usr1_pipe_watch_handler (self, source, condition):
        dprint ("Got USR1 signal, quitting mainloop")
        self.main_loop.quit ()
        return True

    def __usr1_timeout_handler (self):
        dprint ("Timed out waiting for USR1, quitting mainloop")
        self.main_loop.quit ()
        return True

    def __get_xauth_record (self):
        (status, output) = commands.getstatusoutput ("xauth -in list $DISPLAY")
        if status != 0:
            raise XauthParseError, _("'xauth list' returned error")

        for line in output.split ("\n"):
            fields = line.split ("  ")

            # should have display, auth name, auth data
            if len (fields) != 3:
                continue

            # we want the local addr
            if fields[0].find ("/unix") == -1:
                continue

            # hex encoded data, len should be a multiple of 2
            if len (fields[2]) % 2 != 0:
                continue
            
            return (fields[0], fields[1], binascii.unhexlify (fields[2]))
        
        raise XauthParseError, _("'xauth list' returned no records or records in unknown format")

    # Write out a temporary Xauthority file which contains the same magic
    # cookie as the parent display so that we can pass that using -auth
    # to Xnest.
    #
    # Xauthority is a binary file format:
    #
    #       2 bytes         Family value (second byte is as in protocol HOST)
    #       2 bytes         address length (always MSB first)
    #       A bytes         host address (as in protocol HOST)
    #       2 bytes         display "number" length (always MSB first)
    #       S bytes         display "number" string
    #       2 bytes         name length (always MSB first)
    #       N bytes         authorization name string
    #       2 bytes         data length (always MSB first)
    #       D bytes         authorization data string
    #
    def __write_temp_xauth_file (self, wildcard_addr):
        (xauth_display, xauth_name, xauth_data) = self.__get_xauth_record ()

        if wildcard_addr:
            family = 0xffff # FamilyWild
            display_addr = ""
        else:
            family = 0x0100 # FamilyLocal
            display_addr = xauth_display[:xauth_display.find ("/unix")]

        display_num_str = str (self.display_number)
        display_num_len = len (display_num_str)

        display_addr_len = len (display_addr)
        xauth_name_len   = len (xauth_name)
        xauth_data_len   = len (xauth_data)

        pack_format = ">hh%dsh%dsh%dsh%d" % (display_addr_len, display_num_len, xauth_name_len, xauth_data_len)

        blob = struct.pack (pack_format,
                            family,
                            display_addr_len, display_addr,
                            display_num_len, display_num_str,
                            xauth_name_len, xauth_name,
                            xauth_data_len) + xauth_data

        (fd, temp_xauth_file) = tempfile.mkstemp (prefix = ".xauth-")
        os.write (fd, blob)
        os.close (fd)

        dprint ("Wrote temporary xauth file to %s" % temp_xauth_file)

        return temp_xauth_file

    #
    # Need to hold open the first X connection until the session dies
    #
    def __open_x_connection (self, display_name, xauth_file):
        (pipe_r, pipe_w) = os.pipe ()
        pid = os.fork ()
        if pid == 0: # Child process
            os.close (pipe_r)
            new_environ = os.environ
            new_environ["DISPLAY"] = display_name
            new_environ["XAUTHORITY"] = xauth_file
            argv = [ "python", "-c",
                     "import gtk, os, sys; os.write (int (sys.argv[1]), 'Y'); gtk.main ()" ,
                     str (pipe_w) ]
            os.execvpe (argv[0], argv, new_environ)
        os.close (pipe_w)
        select.select ([pipe_r], [], [])[0]
        os.close (pipe_r)

    #
    # FIXME: we have a re-entrancy issue here - if while we're
    # runing the mainloop we re-enter, then we'll install another
    # SIGUSR1 handler and everything will break
    #
    def __start_xnest (self):
        self.display_number = self.__find_free_display ()
        if self.display_number == -1:
            raise SessionStartError, _("Unable to find a free X display")

        self.display_name = ":%d" % self.display_number

        dprint ("Starting Xnest %s" % self.display_name)

        self.xnest_xauth_file = self.__write_temp_xauth_file (True)

        (self.usr1_pipe_r, self.usr1_pipe_w) = os.pipe ()
        self.got_usr1_signal = False
        signal.signal (signal.SIGUSR1, self.__sigusr1_handler)

        self.xnest_pid = os.fork ()
        if self.xnest_pid == 0: # Child process
            signal.signal (signal.SIGUSR1, signal.SIG_IGN)

            argv = XNEST_ARGV + \
                   [ "-auth", self.xnest_xauth_file ] + \
                   [ "-name", _("All Your Settings Are Belong To Us") ]
            argv += [ self.display_name ]

            dprint ("Child process launching %s" % argv)

            os.execv (argv[0], argv)

            # Shouldn't ever reach here
            print "Failed to launch Xnest"
            os._exit (1)

        self.main_loop = gobject.MainLoop ()

        self.xnest_child_watch = gobject.child_watch_add (self.xnest_pid,
                                                          self.__xnest_child_watch_handler)
        io_watch = gobject.io_add_watch (self.usr1_pipe_r,
                                         gobject.IO_IN | gobject.IO_PRI,
                                         self.__usr1_pipe_watch_handler)
        timeout = gobject.timeout_add (ADMIN_TOOL_TIMEOUT * 1000,
                                       self.__usr1_timeout_handler)

        dprint ("Waiting on child process (%d)" % self.xnest_pid)

        self.main_loop.run ()
        self.main_loop = None

        signal.signal (signal.SIGUSR1, signal.SIG_IGN)

        gobject.source_remove (timeout)
        gobject.source_remove (io_watch)
        
        os.close (self.usr1_pipe_r)
        self.usr1_pipe_r = 0
        os.close (self.usr1_pipe_w)
        self.usr1_pipe_w = 0

        if not self.got_usr1_signal:
            if self.xnest_child_watch:
                gobject.source_remove (self.xnest_child_watch)
                self.xnest_child_watch = 0
            
            if self.xnest_pid:
                safe_kill (self.xnest_pid, signal.SIGTERM)
                self.xnest_pid = 0
                raise SessionStartError, _("Failed to start Xnest: timed out waiting for USR1 signal")
            else:
                raise SessionStartError, _("Failed to start Xnest: died during startup")

    def __session_child_watch_handler (self, pid, status):
        dprint ("Session died")
        
        self.session_pid = 0
        self.session_child_watch = 0

        self.force_quit ()

        self.emit ("finished")
            
        return False

    def __prepare_to_run_as_user (self):
        os.setgid (self.user_pw.pw_gid)
        os.setuid (self.user_pw.pw_uid)

        os.chdir (self.temp_homedir)

        # FIXME: setting the selinux context?

        os.setsid ()
        os.umask (022)
        
        new_environ = {}
        for key in PASSTHROUGH_ENVIRONMENT:
            if os.environ.has_key (key):
                new_environ[key] = os.environ[key]

        new_environ["PATH"]       = DEFAULT_PATH 
        new_environ["SHELL"]      = DEFAULT_SHELL
        new_environ["DISPLAY"]    = self.display_name
        new_environ["XAUTHORITY"] = self.session_xauth_file
        new_environ["HOME"]       = self.temp_homedir
        new_environ["LOGNAME"]    = self.user_pw.pw_name
        new_environ["USER"]       = self.user_pw.pw_name
        new_environ["USERNAME"]   = self.user_pw.pw_name

        dprint ("Child process env: %s" % new_environ)

        self.session_xauth_file = None

        return new_environ

    def __start_session (self):
        dprint ("Starting session as %s" % self.user_pw)

        self.session_xauth_file = self.__write_temp_xauth_file (False)
        os.chown (self.session_xauth_file, self.user_pw.pw_uid, self.user_pw.pw_gid)
        
        self.__open_x_connection (self.display_name, self.session_xauth_file)

        self.session_pid = os.fork ()
        if self.session_pid == 0: # Child process
            new_environ = self.__prepare_to_run_as_user ()

            # Apply the profile
            argv = APPLY_TOOL_ARGV + [ self.profile_file ]
            dprint ("Running apply tool: %s" % argv)
            os.spawnve (os.P_WAIT, argv[0], argv, new_environ)

            # Disable sabayon-xinitrc.sh
            new_environ["DISABLE_SABAYON_XINITRC"] = "yes"

            # Start the session
            dprint ("Executing %s" % SESSION_ARGV)
            os.execve (SESSION_ARGV[0], SESSION_ARGV, new_environ)

            # Shouldn't ever happen
            print "Failed to launch Xsession"
            os._exit (1)

        self.session_child_watch = gobject.child_watch_add (self.session_pid,
                                                            self.__session_child_watch_handler)

        # This is totally arbitrary
        self.admin_tool_timeout = gobject.timeout_add (5 * 1000,
                                                       self.__start_admin_tool)

    def __admin_tool_child_watch_handler (self, pid, status):
        dprint ("admin tool died")

        self.admin_tool_pid = 0
        self.admin_tool_child_watch = 0

        self.force_quit ()

        self.emit ("finished")

        return False
    
    def __start_admin_tool (self):
        self.admin_tool_pid = os.fork ()
        if self.admin_tool_pid == 0: # Child process
            new_environ = self.__prepare_to_run_as_user ()

            argv = MONITOR_TOOL_ARGV + [ self.profile_file ]
            
            dprint ("Executing %s" % argv)

            os.execve (argv[0], argv, new_environ)

            # Shouldn't ever reach here
            print "Failed to launch admin tool"
            os._exit (1)

        self.admin_tool_child_watch = gobject.child_watch_add (self.admin_tool_pid,
                                                               self.__admin_tool_child_watch_handler)
        self.admin_tool_timeout = 0
        return False

    def __setup_shell_and_homedir (self):
        dprint ("Setting shell for '%s' to '%s'" % (self.username, DEFAULT_SHELL))
        usermod.set_shell (self.username, DEFAULT_SHELL)

        self.temp_homedir = usermod.create_temporary_homedir (self.user_pw.pw_uid,
                                                              self.user_pw.pw_gid)
        dprint ("Setting temporary home directory for '%s' to '%s'" %
                (self.username, self.temp_homedir))
        usermod.set_homedir (self.username, self.temp_homedir)
        
    def __reset_shell_and_homedir (self):
        dprint ("Unsetting homedir for '%s'" % self.username)
        usermod.set_homedir (self.username, "")

        dprint ("Deleting temporary homedir '%s'" % self.temp_homedir)
        shutil.rmtree (self.temp_homedir)
        
        dprint ("Resetting shell for '%s' to '%s'" % (self.username, NOLOGIN_SHELL))
        usermod.set_shell (self.username, NOLOGIN_SHELL)
        
    def start (self):
        # Set the shell to a runnable one and create an empty homedir
	self.__setup_shell_and_homedir()

        # FIXME: my, what a big hammer you have!
        argv = CLOBBER_USER_PROCESSES_ARGV + [ self.user_pw.pw_name ]
        dprint ("Clobbering existing processes running as user '%s': %s" % (self.user_pw.pw_name, argv))
        os.spawnv (os.P_WAIT, argv[0], argv)
        
        # Get an X server going
        self.__start_xnest ()

        # Start the session as the prototype user
        self.__start_session ()
        
    def force_quit (self):
        self.__kill_admin_tool ()
        self.__kill_session ()
        self.__kill_xnest ()
        self.__reset_shell_and_homedir ()

gobject.type_register (ProtoSession)



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import gtk
import gtk.glade
import userprofile
import util
import aboutdialog
from config import *

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_ADMINTOOL, fmt % args)

class ProfileChangesModel (gtk.ListStore):
    (
        COLUMN_CHANGE,
        COLUMN_IGNORE,
        COLUMN_MANDATORY,
        COLUMN_DESCRIPTION
    ) = range (4)

    def __init__ (self, profile):
        gtk.ListStore.__init__ (self, userprofile.ProfileChange, bool, bool, str, str)

        self.profile = profile
        self.profile.connect ("changed", self.handle_profile_change)
        self.profile.start_monitoring ()

    def handle_profile_change (self, profile, new_change):
        mandatory = False
        ignore = False
        iter = self.get_iter_first ()
        while iter:
            next = self.iter_next (iter)
            change = self[iter][self.COLUMN_CHANGE]
            if change.get_source () == new_change.get_source () and \
               change.get_id ()     == new_change.get_id ():
                ignore    = self[iter][self.COLUMN_IGNORE]
                mandatory = self[iter][self.COLUMN_MANDATORY]
                self.remove (iter)
            iter = next

        row = self.prepend ()
        self.set (row,
                  self.COLUMN_CHANGE,      new_change,
                  self.COLUMN_IGNORE,      ignore,
                  self.COLUMN_MANDATORY,   mandatory,
                  self.COLUMN_DESCRIPTION, new_change.get_short_description ())

class ProfileMonitorWindow:
    #
    # @profile_name: the name of the profile to load/save
    #
    def __init__ (self, profile_name):
        self.profile_name = profile_name
        self.profile      = userprofile.UserProfile (profile_name)
        
        glade_file = os.path.join (GLADEDIR, "sabayon.glade")
        self.xml = gtk.glade.XML (glade_file, "monitor_window")

        self.window = self.xml.get_widget ("monitor_window")
        self.window.connect ("destroy", gtk.main_quit)
        self.window.set_icon_name ("sabayon")
        
        self.treeview = self.xml.get_widget ("changes_treeview")
        self.__setup_treeview ()

        self.save_item = self.xml.get_widget ("save_item")
        self.save_item.connect ("activate", self.__handle_save)
        
        self.quit_item = self.xml.get_widget ("quit_item")
        self.quit_item.connect ("activate", self.__handle_quit)

        self.about_item = self.xml.get_widget ("about_item")
        self.about_item.connect ("activate", self.__handle_about)

        self.window.show ()

    def __handle_save (self, item):
        iter = self.changes_model.get_iter_first ()
        while iter:
            change    = self.changes_model[iter][ProfileChangesModel.COLUMN_CHANGE]
            ignore    = self.changes_model[iter][ProfileChangesModel.COLUMN_IGNORE]
            mandatory = self.changes_model[iter][ProfileChangesModel.COLUMN_MANDATORY]

            if not ignore:
                dprint ("Committing: %s, mandatory = %s" % (change.get_id (), mandatory))
                try:
                    change.get_source ().commit_change (change, mandatory)
                except:
                    util.print_exception ()
                
            iter = self.changes_model.iter_next (iter)
        
        self.changes_model.clear ()
        self.profile.sync_changes ()
    
    def __handle_quit (self, item):
        self.window.destroy ()

    def __handle_about (self, item):
        aboutdialog.show_about_dialog (self.window)
    
    def __on_ignore_toggled (self, toggle, path):
        iter = self.changes_model.get_iter_from_string (path)
        ignore = self.changes_model.get_value (iter, ProfileChangesModel.COLUMN_IGNORE)
        
        ignore = not ignore

        self.changes_model.set (iter, ProfileChangesModel.COLUMN_IGNORE, ignore)
    
    def __on_mandatory_toggled (self, toggle, path):
        iter = self.changes_model.get_iter_from_string (path)
        mandatory = self.changes_model.get_value (iter, ProfileChangesModel.COLUMN_MANDATORY)
        
        mandatory = not mandatory

        self.changes_model.set (iter, ProfileChangesModel.COLUMN_MANDATORY, mandatory)
    
    def __setup_treeview (self):
        self.changes_model = ProfileChangesModel (self.profile)
        self.treeview.set_model (self.changes_model)
        
        self.treeview.get_selection ().set_mode (gtk.SELECTION_SINGLE)
        self.treeview.get_selection ().connect ("changed", self.__treeview_selection_changed)

        toggle = gtk.CellRendererToggle ()
        toggle.connect ("toggled", self.__on_ignore_toggled)
        c = gtk.TreeViewColumn (_("Ignore"),
                                toggle,
                                active = ProfileChangesModel.COLUMN_IGNORE)
        self.treeview.append_column (c)
                                
        toggle = gtk.CellRendererToggle ()
        toggle.connect ("toggled", self.__on_mandatory_toggled)
        c = gtk.TreeViewColumn (_("Mandatory"),
                                toggle,
                                active = ProfileChangesModel.COLUMN_MANDATORY)
        self.treeview.append_column (c)
        
        c = gtk.TreeViewColumn (_("Description"),
                                gtk.CellRendererText (),
                                text = ProfileChangesModel.COLUMN_DESCRIPTION)
        self.treeview.append_column (c)

    def __treeview_selection_changed (self, selection):
        (model, row) = selection.get_selected ()
        if row:
            change = model[row][ProfileChangesModel.COLUMN_CHANGE]
            dprint ("Selected: %s" % model[row][ProfileChangesModel.COLUMN_CHANGE].get_id ())



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import gtk
import util
import userdb

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_ADMINTOOL, fmt % args)
    
class UsersModel (gtk.ListStore):
    (
        COLUMN_USER,
        COLUMN_PROFILE
    ) = range (2)

    def __init__ (self, userdb):
        gtk.ListStore.__init__ (self, str, str)

        self.userdb = userdb

        for username in self.userdb.get_users ():
            profile = self.userdb.get_profile (username)
            if not profile:
                profile = _("None")
            
            self.set (self.append (),
                      self.COLUMN_USER,    username,
                      self.COLUMN_PROFILE, profile)

class ProfilesModel (gtk.ListStore):
    (
        COLUMN_PROFILE,
        COLUMN_BITE_ME_GUIDO
    ) = range (2)

    def __init__ (self, userdb):
        gtk.ListStore.__init__ (self, str)

        self.userdb = userdb

        self.set (self.append (),
                  self.COLUMN_PROFILE, _("None"))
        for profile in self.userdb.get_profiles ():
            self.set (self.append (),
                      self.COLUMN_PROFILE, profile)

class UsersWindow (gtk.Window):
    def __init__ (self, parent_window = None):
        self.userdb = userdb.get_database ()
        
        gtk.Window.__init__ (self, gtk.WINDOW_TOPLEVEL)
        self.set_title (_("All Your Settings Are Belong To Us"))
        self.set_icon_name ("sabayon")
        self.set_transient_for (parent_window)
        self.set_destroy_with_parent (True)
        self.set_default_size (480, 380)

        self.main_vbox = gtk.VBox (False, 0)
        self.main_vbox.show ()
        self.add (self.main_vbox)

        self.scrolled = gtk.ScrolledWindow ()
        self.scrolled.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scrolled.set_shadow_type (gtk.SHADOW_IN)
        self.scrolled.show ()
        self.main_vbox.pack_start (self.scrolled, True, True, 0)

        self.__setup_treeview ()

    def __setup_treeview (self):
        self.users_model = UsersModel (self.userdb)
        self.profiles_model = ProfilesModel (self.userdb)

        self.treeview = gtk.TreeView (self.users_model)
        self.treeview.show ()
        self.scrolled.add (self.treeview)

        self.treeview.get_selection ().set_mode (gtk.SELECTION_SINGLE)
        self.treeview.set_headers_visible (True)

        c = gtk.TreeViewColumn (_("User"),
                                gtk.CellRendererText (),
                                text = UsersModel.COLUMN_USER)
        self.treeview.append_column (c)

        renderer = gtk.CellRendererCombo ()
        renderer.set_property ("text-column", ProfilesModel.COLUMN_PROFILE)
        renderer.set_property ("editable", True)
        renderer.set_property ("has-entry", False)
        renderer.set_property ("model", self.profiles_model)
        renderer.connect ("edited", self.__profile_changed)

        c = gtk.TreeViewColumn (_("Profile"),
                                renderer,
                                text = UsersModel.COLUMN_PROFILE)
        self.treeview.append_column (c)

    def __profile_changed (self, cell, tree_path, profile):
        iter = self.users_model.get_iter_from_string (tree_path)
        username = self.users_model[iter][UsersModel.COLUMN_USER]

        dprint ("Setting profile for '%s' to '%s'", username, profile)
        
        if profile == _("None"):
            self.users_model[iter][UsersModel.COLUMN_PROFILE] = None
        else:
            self.users_model[iter][UsersModel.COLUMN_PROFILE] = profile
        
        self.userdb.set_profile (username, profile)

if __name__ == "__main__":
    util.init_gettext ()

    window = UsersWindow ()
    window.connect ("destroy", gtk.main_quit)
    window.show ()

    gtk.main ()



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import sys
import os.path
import errno
import time
import socket
import shutil
import tempfile
import zipfile
import libxml2
import util
from config import *

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_STORAGE, fmt % args)

def copy_tree (dst_base, src_base, dst_name, src_name = None, overwrite = False):
    if src_name is None:
        src_name = dst_name
    
    try:
        dprint ("Making dir %s" % os.path.join (dst_base, dst_name))
        os.mkdir (os.path.join (dst_base, dst_name))
    except OSError, err:
        if err.errno != errno.EEXIST:
            raise err

    for f in os.listdir (os.path.join (src_base, src_name)):
        src_path = os.path.join (src_base, src_name, f)

        if os.path.isdir (src_path):
            copy_tree (dst_base,
                       src_base,
                       os.path.join (dst_name, f),
                       os.path.join (src_name, f))
        else:
            dst_path = os.path.join (dst_base, dst_name, f)
            if overwrite or not os.path.exists (dst_path):
                shutil.copy2 (src_path, dst_path)

class ProfileStorageException (Exception):
    pass
        
class ProfileStorage:
    """An encapsulation of all the files which make up the
    contents of a profile.

    The files are stored in a ZIP file with metadata. In
    order to add/extract files to the profile, they are first
    copied to/from a temporary directory which is then zipped.
    Note, though, that the fact that its a ZIP file and the
    fact that there is a temporary directory are both
    implementation details and not exposed in the API.

    Profile files are stored in /etc/desktop-profiles.

    Each file or directory in the profile has metadata
    associated with it - the "source" of the file/directory
    and a set of arbitrary key value pairs which that source
    may interpret.

    The profile also contains a revision history, allowing
    you to revert to a previous version of a given file
    or a previous version of the profile itself.
    """
    
    def __init__ (self, name):
        """Create a ProfileStorage.

        @name: the name of the profile - translates to
        /etc/desktop-profiles/$(name).zip.
        """
        self.name = name
        if not os.path.isabs (self.name):
            self.path = os.path.join (PROFILESDIR, self.name + ".zip")
        else:
            self.path = name
            
        self.metadata         = None
        self.zip              = None
        self.temp_path        = None
        self.revisions_path   = None
        self.unsaved_revision = None
        self.needs_saving     = False

        dprint ("Creating profile '%s' from '%s'", self.name, self.path)

    def __del__ (self):
        if self.temp_path:
            shutil.rmtree (self.temp_path)
        self.temp_path = None
        self.revisions_path = None
        
        if self.metadata:
            self.metadata.freeDoc ()
        self.metadata = None
            
        if self.zip:
            self.zip.close ()
        self.zip = None

    def __create_empty_metadata_doc (self):
        metadata = libxml2.newDoc ("1.0")
        root = metadata.newChild (None, "metadata", None)
        root.newChild (None, "profile_revisions", None)
        root.newChild (None, "directories", None)
        root.newChild (None, "files", None)
        return metadata
        
    def __read_metadata (self):
        if not self.metadata is None:
            return

        if not os.path.exists (self.path):
            dprint ("Profile file '%s' doesn't exist" % self.path)
            self.metadata = self.__create_empty_metadata_doc ()
            self.needs_saving = True
            return

        dprint ("Reading metadata from '%s'" % self.path)
            
        try:
            self.zip = zipfile.ZipFile (self.path, "r")
        except:
            raise ProfileStorageException (_("Failed to read file '%s': %s") %
                                           (self.path, sys.exc_info()[1]))

        try:
            blob = self.zip.read ("metadata")
            doc = libxml2.readMemory (blob, len (blob),
                                      "metadata",
                                      None,
                                      libxml2.XML_PARSE_NOBLANKS)
        except:
            raise ProfileStorageException (_("Failed to read metadata from '%s': %s") %
                                           (self.path, sys.exc_info()[1]))
        
        root = doc.getRootElement ()
        if not root or root.name != "metadata":
            doc.freeDoc ()
            raise ProfileStorageException (_("Invalid metadata section in '%s': %s") %
                                           (self.path, sys.exc_info()[1]))
            
        self.metadata = doc

        if len (root.xpathEval ("profile_revisions")) == 0:
            root.newChild (None, "profile_revisions", None)
        if len (root.xpathEval ("directories")) == 0:
            root.newChild (None, "directories", None)
        if len (root.xpathEval ("files")) == 0:
            root.newChild (None, "files", None)

    def __get_profile_revisions (self, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        revisions = []
        for profile_revision in metadata.xpathEval ("/metadata/profile_revisions/profile_revision"):
            revisions.append ((profile_revision.prop ("id"),
                               profile_revision.prop ("timestamp")))
        return revisions
    
    def __get_current_profile_revision (self, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        return metadata.xpathEval ("string(/metadata/profile_revisions/@current)")

    def __get_profile_revision_node (self, profile_revision, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        nodes = metadata.xpathEval ("/metadata/profile_revisions/profile_revision[@id='%s']" % profile_revision)
        if not len (nodes):
            return None
        return nodes[0]

    def __get_profile_revision_item (self, revision, path, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        nodes = metadata.xpathEval ("/metadata/profile_revisions/profile_revision[@id='%s']/item[@path='%s']" %
                                    (revision, path))
        if not len (nodes):
            return None
        return nodes[0]
    
    def __get_profile_revision_items (self, revision, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        return metadata.xpathEval ("/metadata/profile_revisions/profile_revision[@id='%s']/item" % revision)

    def __get_file_revision_node (self, path, revision, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        nodes = metadata.xpathEval ("/metadata/files/file[@path='%s']/revisions/revision[@id='%s']" %
                                    (path, revision))
        if not len (nodes):
            return None
        return nodes[0]
    
    def __get_directory_revision_node (self, path, revision, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        nodes = metadata.xpathEval ("/metadata/directories/directory[@path='%s']/revisions/revision[@id='%s']" %
                                    (path, revision))
        if not len (nodes):
            return None
        return nodes[0]

    def __get_file_revisions (self, path, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        revisions = []
        for revision in metadata.xpathEval ("/metadata/files/file[@path='%s']/revisions/revision" % path):
            revisions.append ((revision.prop ("id"),
                               revision.prop ("timestamp")))
        return revisions
    
    def __get_directory_revisions (self, path, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata
        revisions = []
        for revision in metadata.xpathEval ("/metadata/directories/directory[@path='%s']/revisions/revision" % path):
            revisions.append ((revision.prop ("id"),
                               revision.prop ("timestamp")))
        return revisions

    def __get_revision_source (self, revision_node):
        return revision_node.xpathEval ("string (source)")

    def __get_revision_attributes (self, revision_node):
        attributes = {}
        for node in revision_node.xpathEval ("attributes/attribute"):
            attributes[node.prop ("name")]  = node.prop ("value")
        return attributes

    def __create_new_profile_revision (self, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata

        profile_revisions = metadata.xpathEval ("/metadata/profile_revisions")[0]
        
        current_revision = self.__get_current_profile_revision (metadata)
        if current_revision:
            current_node = self.__get_profile_revision_node (current_revision, metadata)
            new_revision = str (int (current_revision) + 1)
        else:
            current_node = None
            new_revision = "1"
            
        if current_node:
            revision_node = current_node.copyNode (extended = 1)
            current_node.addPrevSibling (revision_node)
        else:
            revision_node = profile_revisions.newChild (None, "profile_revision", None)
            
        revision_node.setProp ("id",        new_revision)
        revision_node.setProp ("timestamp", str (int (time.time ())))
        revision_node.setProp ("user",      util.get_user_name ())
        revision_node.setProp ("host",      socket.gethostname ())

        profile_revisions.setProp ("current", new_revision)

        return new_revision

    def __create_new_file_or_dir_revision (self, file_or_dir_node, source, attributes, metadata):
        # Ensure the revisions node exists
        nodes = file_or_dir_node.xpathEval ("revisions")
        if len (nodes):
            revisions_node = nodes[0]
        else:
            revisions_node = file_or_dir_node.newChild (None, "revisions", None)

        # Find a new revision id
        old_revision_id = 0
        for node in revisions_node.xpathEval ("revision"):
            i = int (node.prop ("id"))
            if i > old_revision_id:
                old_revision_id = i
        revision_id = str (old_revision_id + 1)
        if old_revision_id:
            old_revision_id = str (old_revision_id)
        else:
            old_revision_id = None

        # Create a new revision node
        if revisions_node.children:
            revision_node = metadata.newDocNode (None, "revision", None)
            revisions_node.children.addPrevSibling (revision_node)
        else:
            revision_node = revisions_node.newChild (None, "revision", None)

        # Set properties
        revision_node.setProp ("id",        revision_id)
        revision_node.setProp ("timestamp", str (int (time.time ())))
        revision_node.setProp ("user",      util.get_user_name ())
        revision_node.setProp ("host",      socket.gethostname ())

        # Set source and attributes
        revision_node.newChild (None, "source", source)
        
        attributes_node = revision_node.newChild (None, "attributes", None)
        if attributes:
            for name in attributes:
                attribute_node = attributes_node.newChild (None, "attribute", None)
                attribute_node.setProp ("name",  str (name))
                attribute_node.setProp ("value", str (attributes[name]))

        return (revision_id, old_revision_id)

    def __create_new_file_revision (self, path, source, attributes, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata

        files_node = metadata.xpathEval ("/metadata/files")[0]
        
        # Ensure the file node exists
        nodes = files_node.xpathEval ("file[@path='%s']" % path)
        if len (nodes):
            file_node = nodes[0]
        else:
            file_node = files_node.newChild (None, "file", None)
            file_node.setProp ("path", path)

        return self.__create_new_file_or_dir_revision (file_node, source, attributes, metadata)

    def __create_new_directory_revision (self, path, source, attributes, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata

        files_node = metadata.xpathEval ("/metadata/directories")[0]
        
        # Ensure the directory node exists
        nodes = files_node.xpathEval ("directory[@path='%s']" % path)
        if len (nodes):
            directory_node = nodes[0]
        else:
            directory_node = files_node.newChild (None, "directory", None)
            directory_node.setProp ("path", path)

        return self.__create_new_file_or_dir_revision (directory_node, source, attributes, metadata)

    def __add_profile_revision_item (self, profile_revision, path, type, revision, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata

        profile_revision_node = metadata.xpathEval ("/metadata/profile_revisions/profile_revision[@id='%s']" %
                                                    profile_revision)[0]

        nodes = profile_revision_node.xpathEval ("item[@path='%s']" % path)
        if len (nodes):
            item_node = nodes[0]
        else:
            item_node = profile_revision_node.newChild (None, "item", None)
            item_node.setProp ("path", path)

        item_node.setProp ("type", type)
        item_node.setProp ("revision", revision)

    def __item_revision_is_current (self, path, type, revision, metadata = None):
        if metadata is None:
            metadata = self.metadata
        assert metadata

        current_profile_revision = self.__get_current_profile_revision (metadata)
        if not current_profile_revision:
            return False

        return metadata.xpathEval ("boolean(/metadata/profile_revisions/profile_revision[@id='%s']"
                                   "/item[@path='%s' and @type='%s' and @revision='%s'])" %
                                   (current_profile_revision, path, type, revision))
        
    def __unpack (self):
        self.__read_metadata ()

        if self.temp_path:
            return
        
        self.temp_path = tempfile.mkdtemp (prefix = "sabayon-profile-storage-")
        self.revisions_path = os.path.join (self.temp_path, "%revisions%")
        os.mkdir (self.revisions_path)

        dprint ("Unpacking '%s' in '%s'" % (self.path, self.temp_path))

        if not self.zip:
            return
        
        def unzip_directory (zip, dir, name):
            if not os.path.exists (os.path.join (dir, name)):
                os.makedirs (os.path.join (dir, name))
            for f in zip.namelist ():
                if not f.startswith (name):
                    continue

                dest_path = os.path.join (dir, f)
                dest_dir  = os.path.dirname (dest_path)

                if not os.path.exists (dest_dir):
                    os.makedirs (dest_dir)
                
                # It sucks that we lose file permissions, mtime etc. with ZIP
                file (dest_path, "w").write (zip.read (f))
        
        def unzip_foreach (path, revision, is_current, is_directory, data):
            (zip, temp_path, revisions_path) = data

            dprint ("Unzip: path = %s, revision = %s, is_current = %s, is_directory = %s",
                    path, revision, is_current, is_directory)
            
            if is_current:
                abs_path = os.path.join (temp_path, path)
            else:
                abs_path = os.path.join (revisions_path, path, revision)
                path = os.path.join ("%revisions%", path, revision)
                
            if is_directory:
                unzip_directory (zip, temp_path, path)
            else:
                dest_dir = os.path.join (temp_path, os.path.dirname (path))
                if not os.path.exists (dest_dir):
                    os.makedirs (dest_dir)
                                                           
                # It sucks that we lose file permissions, mtime etc. with ZIP
                file (abs_path, "w").write (zip.read (path))

        self.__foreach_all (unzip_foreach, (self.zip, self.temp_path, self.revisions_path))

    def __foreach_revision (self, node, type, callback, user_data):
        path = node.prop ("path")
        for revision in node.xpathEval ("revisions/revision"):
            revision_id = revision.prop ("id")
            callback (path,
                      revision_id,
                      self.__item_revision_is_current (path, type, revision_id),
                      type == "directory",
                      user_data)

    def __foreach_all (self, callback, user_data):
        for file_node in self.metadata.xpathEval ("/metadata/files/file"):
            self.__foreach_revision (file_node, "file", callback, user_data)
        for directory_node in self.metadata.xpathEval ("/metadata/directories/directory"):
            self.__foreach_revision (directory_node, "directory", callback, user_data)

    def copy (self, name):
        """Create a new ProfileStorage object, copying the
        contents of this profile to the new profile.

        @name: the name of the new profile.

        Return value: a #ProfileStorage object.
        """
        if os.path.isfile (self.path) and zipfile.is_zipfile (self.path):
            new_path = os.path.join (PROFILESDIR, name + ".zip")
            try:
                shutil.copyfile (self.path, new_path)
            except:
                dprint ("Failed to copy profile from '%s' to '%s': %s" % \
                        (self.path, new_path, sys.exc_info()[1]))
        
        retval = ProfileStorage (name)
        retval.clear_revisions ()
        retval.save ()
        return retval

    def add (self, path, src_dir, source, attributes = None):
        """Add a new - or update an existing - file or directory
        to the profile. If @path is a directory, then the contents
        of the directory will be recursively saved in the profile.

        @path: the relative path of the file or directory.
        @src_dir: the directory (which @path is relative to) from
        which the file or directory contents should be copied to
        the profile.
        @source: the name of the #ProfileSource with which the file
        or directory is associated.
        @attributes: a varargs list of arbitrary key/value pairs
        (specific to @source) which should be saved.
        """
        dprint ("Adding '%s' from %s:%s" % (path, source, src_dir))
        
        self.__unpack ()

        src_path = os.path.join (src_dir, path)
        dst_path = os.path.join (self.temp_path, path)

        if not os.path.exists (src_path):
            raise ProfileStorageException (_("Cannot add non-existent file '%s'") % src_path)
        
        if not self.unsaved_revision:
            self.unsaved_revision = self.__create_new_profile_revision ()
            
        if os.path.isdir (src_path):
            (new_revision, old_revision) = self.__create_new_directory_revision (path, source, attributes)
            if old_revision:
                needs_copying = self.__item_revision_is_current (path, "directory", old_revision)
            else:
                needs_copying = False
            self.__add_profile_revision_item (self.unsaved_revision, path, "directory", new_revision)

            if needs_copying:
                revision_path = os.path.join (self.revisions_path, path, old_revision)
                dprint ("Retaining old revision of '%s' from '%s' to '%s'",
                        path, dst_path, revision_path)
                os.renames (dst_path, revision_path)
            
            copy_tree (self.temp_path, src_dir, path)
        else:
            (new_revision, old_revision) = self.__create_new_file_revision (path, source, attributes)
            if old_revision:
                needs_copying = self.__item_revision_is_current (path, "file", old_revision)
            else:
                needs_copying = False
            self.__add_profile_revision_item (self.unsaved_revision, path, "file", new_revision)
            
            if needs_copying:
                revision_path = os.path.join (self.revisions_path, path, old_revision)
                dprint ("Retaining old revision of '%s' from '%s' to '%s'",
                        path, dst_path, revision_path)
                os.renames (dst_path, revision_path)

            dirname = os.path.dirname (dst_path)
            if not os.path.exists (dirname):
                os.makedirs (dirname)
            shutil.copy2 (src_path, dst_path)

        self.needs_saving = True

    def remove (self, path):
        """Remove a file or directory from the profile.

        @path: the relative path of the file or directory. This is
        the same path used with ProfileStorage::add().
        """
        dprint ("Removing '%s' profile from profile %s", path, self.name)

        self.__unpack ()
        
        if not self.unsaved_revision:
            self.unsaved_revision = self.__create_new_profile_revision ()

        item_node = self.__get_profile_revision_item (self.unsaved_revision, path)
        if not item_node:
            return

        item_type     = item_node.prop ("type")
        item_revision = item_node.prop ("revision")

        item_node.unlinkNode ()
        item_node.freeNode ()
        
        os.renames (os.path.join (self.temp_path, path),
                    os.path.join (self.revisions_path, path, item_revision))

        self.needs_saving = True
        
    def extract (self, path, dst_dir, overwrite = False, revision = None):
        """Extract a file or directory from the profile.

        @path: the relative path of the file or directory to extract.
        This is the same path used with ProfileStorage::add().
        @dst_dir: the directory to which the file or directory specified
        by @path should be extracted. Any subdirectories of @dst_dir
        specified by @path will be created.
        @revision: an (optional) revision identifier which specifies
        the revision of the file or directory which should be extracted.
        The revision identifier must have been returned from
        ProfileStorage::get_revisions() and may be a profile or file
        revision.
        """
        dprint ("Extracting '%s' to '%s', revision %s" % (path, dst_dir, revision))
        
        self.__unpack ()

        if not revision:
            profile_revision = self.__get_current_profile_revision ()
            if not profile_revision:
                raise ProfileStorageException (_("No current revision"))
            
            item = self.__get_profile_revision_item (profile_revision, path)
            if not item:
                raise ProfileStorageException (_("'%s' does not exist in profile revision '%s'"),
                                               path, profile_revision)

            item_type     = item.prop ("type")
            item_revision = item.prop ("revision")
        else:
            (item_type, item_revision) = revision.split (":")

            if item_type == "profile":
                item = self.__get_profile_revision_item (item_revision, path)
                if not item:
                    raise ProfileStorageException (_("'%s' does not exist in profile revision '%s'"),
                                                   path, profile_revision)
                    return

                item_type     = item.prop ("type")
                item_revision = item.prop ("revision")

        if self.__item_revision_is_current (path, item_type, item_revision):
            if item_type == "directory":
                copy_tree (dst_dir, self.temp_path, path, None, overwrite)
            else:
                dst_path = os.path.join (dst_dir, path)
                if overwrite or not os.path.exists (dst_path):
                    dirname = os.path.dirname (dst_path)
                    if not os.path.exists (dirname):
                        os.makedirs (dirname)
                    shutil.copy2 (os.path.join (self.temp_path, path), dst_path)
        else:
            if item_type == "directory":
                copy_tree (dst_dir,
                           os.path.join (self.revisions_path, path),
                           path,
                           item_revision,
                           overwrite)
            else:
                dst_path = os.path.join (dst_dir, path)
                if overwrite or not os.path.exists (dst_path):
                    dirname = os.path.dirname (dst_path)
                    if not os.path.exists (dirname):
                        os.makedirs (dirname)
                    shutil.copy2 (os.path.join (self.revisions_path, path, item_revision), dst_path)

    def list (self, source = None, profile_revision = None):
        """List the current contents of the profile.

        @source: an (optional) identifier of the source whose
        files should be listed. This is identifier as @source
        passed to ProfileStorage::add().
        @profile_revision: specify the profile revision whose contents
        should be listed. Defaults to the current revision.

        Return value: a list of (@source, @path) tuples.
        """
        def listify (source, path, retval):
            retval.append ((source, path))

        retval = []
        self.foreach (listify, retval, source, profile_revision)

        return retval

    def foreach (self, callback, user_data = None, source = None, profile_revision = None):
        """Iterate over the contents of the profile:

        @callback: an function or method of which takes 
        at least two arguments - @source and @path. If @callback
        is a method, the object which the method is associated
        will be the first parameter. If @user_data is passed,
        then it will be the final parameter to the callback.
        @user_data: an (optional) parameter to pass to @callback.
        @source: an (optional) identifier of the source whose
        files should be listed.
        @profile_revision: an (optional) profile revision from
        which the files should be listed. Defaults to the current
        revision.
        """
        self.__read_metadata ()

        if not profile_revision:
            profile_revision = self.__get_current_profile_revision ()
        else:
            if not profile_revision.startswith ("profile:"):
                raise ProfileStorageException (_("Not a valid profile revision"))
            profile_revision = profile_revision[len ("profile:"):]
        
        if not profile_revision:
            return

        for item in self.__get_profile_revision_items (profile_revision):
            item_path     = item.prop ("path")
            item_type     = item.prop ("type")
            item_revision = item.prop ("revision")

            if item_type == "file":
                revision_node = self.__get_file_revision_node (item_path, item_revision)
            elif item_type == "directory":
                revision_node = self.__get_directory_revision_node (item_path, item_revision)
            else:
                dprint ("Unknown item type '%s' for path '%s' in revision '%s'" %
                        (item_type, item_path, item_revision))
                continue

            if not revision_node:
                dprint ("No revision '%s' for %s '%s" % (item_revision, item_type, item_path))
                continue

            item_source = self.__get_revision_source (revision_node)
            if not item_source:
                dprint ("No source associated with item '%s'" % item_path)
                continue

            if source and source != item_source:
                continue

            if not user_data is None:
                callback (item_source, item_path, user_data)
            else:
                callback (item_source, item_path)

    def save (self):
        """Save the contents of the profile to
        /etc/desktop-profiles/$(name).zip.
        """
        self.__read_metadata ()
        if not self.needs_saving:
            dprint ("No changes to profile '%s' need saving" % self.name)
            return

        def failsafe_rename (src, dst):
            try:
                os.rename (src, dst)
            except:
                pass

        if os.path.exists (self.path):
            backup = self.path + ".bak"
            failsafe_rename (self.path, backup)
        else:
            backup = None
        
        dprint ("Writing contents of profile to '%s'" % self.path)
        
        try:
            save_zip = zipfile.ZipFile (self.path, "w")

            save_zip.writestr ("metadata", self.metadata.serialize (format = 1))
            self.unsaved_revision = None

            def zip_directory (save_zip, dir, name):
                for f in os.listdir (dir):
                    path = os.path.join (dir, f)
                    if os.path.isdir (path):
                        zip_directory (save_zip,
                                       path,
                                       os.path.join (name, f))
                    elif os.path.isfile (path):
                        save_zip.write (path, os.path.join (name, f))
        
            def zip_foreach (path, revision, is_current, is_directory, data):
                (save_zip, temp_path, revisions_path) = data

                if is_current:
                    abs_path = os.path.join (temp_path, path)
                else:
                    abs_path = os.path.join (revisions_path, path, revision)
                    path = os.path.join ("%revisions%", path, revision)
                    
                if is_directory:
                    zip_directory (save_zip, abs_path, path)
                else:
                    save_zip.write (abs_path, path)

            self.__foreach_all (zip_foreach, (save_zip, self.temp_path, self.revisions_path))

            save_zip.close ()
        except:
            if backup:
                failsafe_rename (backup, self.path)
            raise

        if backup:
            os.remove (backup)

        if self.temp_path:
            shutil.rmtree (self.temp_path)
            self.temp_path = None
            self.revisions_path = None
        
        self.needs_saving = False
        
        self.zip = zipfile.ZipFile (self.path, "r")

    def __get_revision_node (self, path, revision):
        self.__read_metadata ()

        if not revision:
            profile_revision = self.__get_current_profile_revision ()
            if not profile_revision:
                return None
            
            item = self.__get_profile_revision_item (profile_revision, path)
            if not item:
                return None

            item_type     = item.prop ("type")
            item_revision = item.prop ("revision")
        else:
            (item_type, item_revision) = revision.split (":")

            if item_type == "profile":
                item = self.__get_profile_revision_item (item_revision, path)
                if not item:
                    return None

                item_type     = item.prop ("type")
                item_revision = item.prop ("revision")

        if item_type == "file":
            return self.__get_file_revision_node (path, item_revision)
        elif item_type == "directory":
            return self.__get_directory_revision_node (path, item_revision)

        return None

    def get_attributes (self, path, revision = None):
        """Return the attributes associated with a file or directory
        from the profile.

        @path: the relative path of the file or directory to look up.
        This is the same path used with ProfileStorage::add().
        @revision: the profile or file/directory revision identifier
        with which to look up.

        Return value: a dictionary containing the key/value pairs
        passed to ProfileStorage::add().
        """
        revision_node = self.__get_revision_node (path, revision)
        if revision_node is None:
            return {}
        return self.__get_revision_attributes (revision_node)

    def get_source (self, path, revision = None):
        """Return the source associated with a file or directory
        from the profile.

        @path: the relative path of the file or directory to look up.
        This is the same path used with ProfileStorage::add().
        @revision: the profile or file/directory revision identifier
        with which to look up.

        Return value: a source identifier.
        """
        revision_node = self.__get_revision_node (path, revision)
        if revision_node is None:
            return None
        return self.__get_revision_source (revision_node)

    def get_revisions (self, path = None):
        """Retrieve the list of profile revisions or the list of
        revisions of a given file or directory.
        
        @path: the relative path of the file or directory to look up.
        This is the same path used with ProfileStorage::add().

        Return value: a list of revision identifiers and timestamps
        for the profile or file/directory in chronological order.
        """
        self.__read_metadata ()

        revisions = []
        if not path:
            for (revision, timestamp) in self.__get_profile_revisions ():
                revisions.append (("profile:%s" % revision, timestamp))
        else:
            for (revision, timestamp) in self.__get_file_revisions (path):
                revisions.append (("file:%s" % revision, timestamp))
            for (revision, timestamp) in self.__get_directory_revisions (path):
                revisions.append (("directory:%s" % revision, timestamp))

        return revisions

    def __copy_to_new_metadata_foreach (self, source, path, data):
        (metadata, unsaved_revision) = data
        
        attributes = self.get_attributes (path)

        if os.path.isdir (os.path.join (self.temp_path, path)):
            (new_revision, old_revision) = self.__create_new_directory_revision (path, source, attributes, metadata)
            assert not old_revision
            self.__add_profile_revision_item (unsaved_revision, path, "directory", new_revision, metadata)
        else:
            (new_revision, old_revision) = self.__create_new_file_revision (path, source, attributes, metadata)
            assert not old_revision
            self.__add_profile_revision_item (unsaved_revision, path, "file", new_revision, metadata)
        
    def clear_revisions (self):
        """Remove all profile and file/directory revision history."""
        dprint ("Clearing revision history from profile '%s'", self.name)

        self.__unpack ()

        metadata = self.__create_empty_metadata_doc ()
        unsaved_revision = self.__create_new_profile_revision (metadata)

        self.foreach (self.__copy_to_new_metadata_foreach, (metadata, unsaved_revision))

        shutil.rmtree (self.revisions_path)
        os.mkdir (self.revisions_path)

        self.metadata.freeDoc ()
        self.metadata = metadata
        self.unsaved_revision = unsaved_revision
        self.needs_saving = True

    def revert (self, revision, path = None):
        """Revert @profile, or a specific @path in @profile to
        a given @revision.

        @revision: the profile, file or directory revision to revert
        @profile (or @path, if given) to. A file or directory revision
        may only be used if @path is given.
        @path: the (optional) path of the file or directory which
        should be reverted.
        """
        if path:
            dprint ("Reverting '%s' to revision '%s'", path, revision)
        else:
            dprint ("Reverting profile '%s' to revision '%s'", self.name, revision)
        
        self.__unpack ()
        
        extract_dir = tempfile.mkdtemp (prefix = "sabayon-profile-storage-")
        
        (revision_type, revision_number) = revision.split (":")
        if revision_type == "profile":
            # Create new revision and remove everything
            def remove_from_current_foreach (source, path, self):
                self.remove (path)
            self.foreach (remove_from_current_foreach, self)

            # Add everything from @revision
            def add_from_revision_foreach (source, path, data):
                (self, extract_dir) = data
                self.extract (path, extract_dir, revision = revision)
                attributes = self.get_attributes (path, revision)
                self.add (path, extract_dir, source, attributes)
            self.foreach (add_from_revision_foreach,
                          (self, extract_dir),
                          profile_revision = revision)
        else:
            assert path
            self.extract (path, extract_dir, revision = revision)
            attributes = self.get_attributes (path, revision)
            source = self.get_source (path, revision)
            self.add (path, extract_dir, source, attributes)
            
        shutil.rmtree (extract_dir)

#
# Unit tests
#
def run_unit_tests ():
    # Remove cruft
    profile_path = os.path.join (os.getcwd (), "storage-test.zip")
    if os.path.exists (profile_path):
        os.remove (profile_path)

    # Create temporary dir for test
    temp_dir = tempfile.mkdtemp (prefix = "storage-test-")

    # Create profile
    profile = ProfileStorage (profile_path)
    
    # Save the profile (no revision yet)
    profile.save ()
    assert os.path.exists (profile_path)

    # Create two test files in the temporary dir
    os.mkdir (os.path.join (temp_dir, "t"))
    open (os.path.join (temp_dir, "t/config1.test"), "w").write ("new test file 1")
    open (os.path.join (temp_dir, "config2.test"), "w").write ("new test file 2")

    # Add the two test files to the profile
    profile.add ("t/config1.test", temp_dir, "TestSource1", { "foo-attr1" : "foo" , "bar-attr1" : 1 })
    profile.add ("config2.test", temp_dir, "TestSource2", { "foo-attr2" : "foo" , "bar-attr2" : 2 })

    # Create a test directory with some files
    os.makedirs (os.path.join (temp_dir, "foobar/foo/bar/foo/bar"))
    open (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml"), "w").write ("new test file 3")
    open (os.path.join (temp_dir, "foobar/foo/bar/foo.txt"), "w").write ("new test file 4")
    open (os.path.join (temp_dir, "foobar/foo/bar.txt"), "w").write ("new test file 5")
    open (os.path.join (temp_dir, "foobar/foo.txt"), "w").write ("new test file 6")

    # Add the test directory to the profile
    profile.add ("foobar", temp_dir, "TestSource3", { "foo-attr3" : "foo" , "bar-attr3" : 3 })
    
    # Save the profile (first revision)
    profile.save ()
    assert os.path.exists (profile_path)
    
    # Add one of the test files again to get a new revision
    open (os.path.join (temp_dir, "t/config1.test"), "w").write ("new test file 99")
    profile.add ("t/config1.test", temp_dir, "TestSource99", { "foo-attr99" : "foo" , "bar-attr99" : 99 })

    # Remove one of the test files
    profile.remove ("config2.test")
    
    # Add the test directory again to get a new revision
    open (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml"), "w").write ("new test file 2005")
    open (os.path.join (temp_dir, "foobar/foo/bar/foo.txt"), "w").write ("new test file 2005")
    open (os.path.join (temp_dir, "foobar/foo/bar.txt"), "w").write ("new test file 2005")
    open (os.path.join (temp_dir, "foobar/foo.txt"), "w").write ("new test file 2005")
    profile.add ("foobar", temp_dir, "TestSource2005", { "foo-attr2005" : "foo" , "bar-attr2005" : 2005 })

    # Add another directory
    os.makedirs (os.path.join (temp_dir, "blaas/are/nice"))
    open (os.path.join (temp_dir, "blaas/are/nice/foo.txt"), "w").write ("blaas are nice")
    profile.add ("blaas", temp_dir, "Waterford", { "nice" : True, "with-butter" : "but of course" })
    
    # Remove temporary dir
    shutil.rmtree (temp_dir)
    
    # Save the profile (second revision)
    os.remove (profile_path)
    profile.save ()
    assert os.path.exists (profile_path)

    # Create temporary dir again
    temp_dir = tempfile.mkdtemp (prefix = "storage-test-")
    
    # Now, re-open
    profile = ProfileStorage (profile_path)

    # Remove a directory
    profile.remove ("blaas")

    # Re-add one of the files
    open (os.path.join (temp_dir, "config2.test"), "w").write ("I'm back, yes its me!")
    profile.add ("config2.test", temp_dir, "TestSource2")

    # Save the profile (third revision)
    os.remove (profile_path)
    profile.save ()
    assert os.path.exists (profile_path)

    # Remove temp dir again
    shutil.rmtree (temp_dir)

    # Now, re-open and validate
    profile = ProfileStorage (profile_path)

    # We should have three revisions, we saved thrice
    profile_revisions = profile.get_revisions ()
    assert len (profile_revisions) == 3
    assert profile_revisions[0][0] == "profile:3"
    assert profile_revisions[1][0] == "profile:2"
    assert profile_revisions[2][0] == "profile:1"

    # Verify the latest revision
    l = profile.list ()
    assert len (l) == 3

    (source, path) = l[0]
    assert source == "TestSource99"
    assert path == "t/config1.test"
    assert profile.get_source (path) == "TestSource99"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr99")
    assert attributes["foo-attr99"] == "foo"
    assert attributes.has_key ("bar-attr99")
    assert attributes["bar-attr99"] == "99"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 2
    assert revisions[0][0] == "file:2"
    assert revisions[1][0] == "file:1"

    (source, path) = l[1]
    assert source == "TestSource2005"
    assert path == "foobar"
    assert profile.get_source (path) == "TestSource2005"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr2005")
    assert attributes["foo-attr2005"] == "foo"
    assert attributes.has_key ("bar-attr2005")
    assert attributes["bar-attr2005"] == "2005"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 2
    assert revisions[0][0] == "directory:2"
    assert revisions[1][0] == "directory:1"

    (source, path) = l[2]
    assert source == "TestSource2"
    assert path == "config2.test"
    assert profile.get_source (path) == "TestSource2"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 0
    revisions = profile.get_revisions (path)
    assert len (revisions) == 2
    assert revisions[0][0] == "file:2"
    assert revisions[1][0] == "file:1"
    
    # Create temporary dir for extraction
    temp_dir = tempfile.mkdtemp (prefix = "storage-test-")

    # Extract each of the files/directories
    for (source, path) in l:
        profile.extract (path, temp_dir)

    # Verify their contents
    assert os.path.isfile (os.path.join (temp_dir, "t/config1.test"))
    assert file (os.path.join (temp_dir, "t/config1.test")).read () == "new test file 99"
    assert os.path.isfile (os.path.join (temp_dir, "config2.test"))
    assert file (os.path.join (temp_dir, "config2.test")).read () == "I'm back, yes its me!"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo.txt")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar.txt")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo.txt")).read () == "new test file 2005"
    
    # Remove temporary extraction dir
    shutil.rmtree (temp_dir)


    # Verify the second revision
    l = profile.list (profile_revision = profile_revisions[1][0])
    assert len (l) == 3

    (source, path) = l[0]
    assert source == "TestSource99"
    assert path == "t/config1.test"
    assert profile.get_source (path) == "TestSource99"
    attributes = profile.get_attributes (path, profile_revisions[1][0])
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr99")
    assert attributes["foo-attr99"] == "foo"
    assert attributes.has_key ("bar-attr99")
    assert attributes["bar-attr99"] == "99"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 2
    assert revisions[0][0] == "file:2"
    assert revisions[1][0] == "file:1"

    # Try using one of the file revisions
    assert profile.get_source (path, revisions[1][0]) == "TestSource1"
    attributes = profile.get_attributes (path, revisions[1][0])
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr1")
    assert attributes["foo-attr1"] == "foo"
    assert attributes.has_key ("bar-attr1")
    assert attributes["bar-attr1"] == "1"
    
    (source, path) = l[1]
    assert source == "TestSource2005"
    assert path == "foobar"
    assert profile.get_source (path, profile_revisions[1][0]) == "TestSource2005"
    attributes = profile.get_attributes (path, profile_revisions[1][0])
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr2005")
    assert attributes["foo-attr2005"] == "foo"
    assert attributes.has_key ("bar-attr2005")
    assert attributes["bar-attr2005"] == "2005"
    
    (source, path) = l[2]
    assert source == "Waterford"
    assert path == "blaas"
    assert profile.get_source (path, profile_revisions[1][0]) == "Waterford"
    attributes = profile.get_attributes (path, profile_revisions[1][0])
    assert len (attributes) == 2
    assert attributes.has_key ("with-butter")
    assert attributes["with-butter"] == "but of course"
    assert attributes.has_key ("nice")
    assert attributes["nice"] == "True"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 1
    assert revisions[0][0] == "directory:1"

    # Create temporary dir for extraction
    temp_dir = tempfile.mkdtemp (prefix = "storage-test-")

    # Extract each of the files/directories
    for (source, path) in l:
        profile.extract (path, temp_dir, revision = profile_revisions[1][0])

    # Verify their contents
    assert os.path.isfile (os.path.join (temp_dir, "t/config1.test"))
    assert file (os.path.join (temp_dir, "t/config1.test")).read () == "new test file 99"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo.txt")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar.txt")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo.txt")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "blaas/are/nice/foo.txt"))
    assert file (os.path.join (temp_dir, "blaas/are/nice/foo.txt")).read () == "blaas are nice"
    
    # Remove temporary extraction dir
    shutil.rmtree (temp_dir)

    
    # Verify the first revision
    l = profile.list (profile_revision = profile_revisions[2][0])
    assert len (l) == 3

    (source, path) = l[0]
    assert source == "TestSource1"
    assert path == "t/config1.test"
    assert profile.get_source (path, profile_revisions[2][0]) == "TestSource1"
    attributes = profile.get_attributes (path, profile_revisions[2][0])
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr1")
    assert attributes["foo-attr1"] == "foo"
    assert attributes.has_key ("bar-attr1")
    assert attributes["bar-attr1"] == "1"
    
    (source, path) = l[1]
    assert source == "TestSource2"
    assert path == "config2.test"
    assert profile.get_source (path, profile_revisions[2][0]) == "TestSource2"
    attributes = profile.get_attributes (path, profile_revisions[2][0])
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr2")
    assert attributes["foo-attr2"] == "foo"
    assert attributes.has_key ("bar-attr2")
    assert attributes["bar-attr2"] == "2"
    
    (source, path) = l[2]
    assert source == "TestSource3"
    assert path == "foobar"
    assert profile.get_source (path, profile_revisions[2][0]) == "TestSource3"
    attributes = profile.get_attributes (path, profile_revisions[2][0])
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr3")
    assert attributes["foo-attr3"] == "foo"
    assert attributes.has_key ("bar-attr3")
    assert attributes["bar-attr3"] == "3"

    # Create temporary dir for extraction
    temp_dir = tempfile.mkdtemp (prefix = "storage-test-")

    # Extract each of the files/directories
    for (source, path) in l:
        profile.extract (path, temp_dir, revision = profile_revisions[2][0])

    # Verify their contents
    assert os.path.isfile (os.path.join (temp_dir, "t/config1.test"))
    assert file (os.path.join (temp_dir, "t/config1.test")).read () == "new test file 1"
    assert os.path.isfile (os.path.join (temp_dir, "config2.test"))
    assert file (os.path.join (temp_dir, "config2.test")).read () == "new test file 2"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml")).read () == "new test file 3"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo.txt")).read () == "new test file 4"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar.txt")).read () == "new test file 5"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo.txt")).read () == "new test file 6"
    
    # Remove temporary extraction dir
    shutil.rmtree (temp_dir)

    # Test reverting
    profile = ProfileStorage (profile_path)
    profile_revisions = profile.get_revisions ()
    assert len (profile_revisions) == 3
    assert profile_revisions[0][0] == "profile:3"
    assert profile_revisions[1][0] == "profile:2"
    assert profile_revisions[2][0] == "profile:1"

    # Revert whole thing to second revision
    profile.revert (profile_revisions[1][0])

    # Save
    os.remove (profile_path)
    profile.save ()
    assert os.path.exists (profile_path)

    # Open the reverted profile
    profile = ProfileStorage (profile_path)
    profile_revisions = profile.get_revisions ()
    assert len (profile_revisions) == 4
    assert profile_revisions[0][0] == "profile:4"
    assert profile_revisions[1][0] == "profile:3"
    assert profile_revisions[2][0] == "profile:2"
    assert profile_revisions[3][0] == "profile:1"

    # Verify the latest revision
    l = profile.list ()
    assert len (l) == 3

    (source, path) = l[0]
    assert source == "TestSource99"
    assert path == "t/config1.test"
    assert profile.get_source (path) == "TestSource99"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr99")
    assert attributes["foo-attr99"] == "foo"
    assert attributes.has_key ("bar-attr99")
    assert attributes["bar-attr99"] == "99"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 3
    assert revisions[0][0] == "file:3"
    assert revisions[1][0] == "file:2"
    assert revisions[2][0] == "file:1"

    (source, path) = l[1]
    assert source == "TestSource2005"
    assert path == "foobar"
    assert profile.get_source (path) == "TestSource2005"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr2005")
    assert attributes["foo-attr2005"] == "foo"
    assert attributes.has_key ("bar-attr2005")
    assert attributes["bar-attr2005"] == "2005"
    
    (source, path) = l[2]
    assert source == "Waterford"
    assert path == "blaas"
    assert profile.get_source (path) == "Waterford"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("with-butter")
    assert attributes["with-butter"] == "but of course"
    assert attributes.has_key ("nice")
    assert attributes["nice"] == "True"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 2
    assert revisions[0][0] == "directory:2"
    assert revisions[1][0] == "directory:1"

    # Clear revision history
    profile.clear_revisions ()

    # Save
    os.remove (profile_path)
    profile.save ()
    assert os.path.exists (profile_path)

    # We should only have a single revision now
    profile_revisions = profile.get_revisions ()
    assert len (profile_revisions) == 1
    assert profile_revisions[0][0] == "profile:1"

    # Verify this revision
    l = profile.list ()
    assert len (l) == 3

    (source, path) = l[0]
    assert source == "TestSource99"
    assert path == "t/config1.test"
    assert profile.get_source (path) == "TestSource99"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr99")
    assert attributes["foo-attr99"] == "foo"
    assert attributes.has_key ("bar-attr99")
    assert attributes["bar-attr99"] == "99"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 1
    assert revisions[0][0] == "file:1"

    (source, path) = l[1]
    assert source == "TestSource2005"
    assert path == "foobar"
    assert profile.get_source (path) == "TestSource2005"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("foo-attr2005")
    assert attributes["foo-attr2005"] == "foo"
    assert attributes.has_key ("bar-attr2005")
    assert attributes["bar-attr2005"] == "2005"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 1
    assert revisions[0][0] == "directory:1"

    (source, path) = l[2]
    assert source == "Waterford"
    assert path == "blaas"
    assert profile.get_source (path) == "Waterford"
    attributes = profile.get_attributes (path)
    assert len (attributes) == 2
    assert attributes.has_key ("with-butter")
    assert attributes["with-butter"] == "but of course"
    assert attributes.has_key ("nice")
    assert attributes["nice"] == "True"
    revisions = profile.get_revisions (path)
    assert len (revisions) == 1
    assert revisions[0][0] == "directory:1"
    
    # Create temporary dir for extraction
    temp_dir = tempfile.mkdtemp (prefix = "storage-test-")

    # Extract each of the files/directories
    for (source, path) in l:
        profile.extract (path, temp_dir)

    # Verify their contents
    assert os.path.isfile (os.path.join (temp_dir, "t/config1.test"))
    assert file (os.path.join (temp_dir, "t/config1.test")).read () == "new test file 99"
    assert os.path.isfile (os.path.join (temp_dir, "blaas/are/nice/foo.txt"))
    assert file (os.path.join (temp_dir, "blaas/are/nice/foo.txt")).read () == "blaas are nice"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo/bar/%gconf.xml")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar/foo.txt")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo/bar.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo/bar.txt")).read () == "new test file 2005"
    assert os.path.isfile (os.path.join (temp_dir, "foobar/foo.txt"))
    assert file (os.path.join (temp_dir, "foobar/foo.txt")).read () == "new test file 2005"
    
    # Remove temporary extraction dir
    shutil.rmtree (temp_dir)
    
    os.remove (profile_path)



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import gconf
import gconfsource

try:
    import userprofile
    import util
    from config import *
except:
    from sabayon import userprofile
    from sabayon import util
    from sabayon.config import *

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_PANELDELEGATE, fmt % args)

class PanelChange (userprofile.ProfileChange):
    def __init__ (self, source, delegate, id):
        userprofile.ProfileChange.__init__ (self, source, delegate)
        self.id = id
    def get_id (self):
        return self.id

#
# FIXME: these short descriptions are lame, should be:
#  "Battery applet added to top panel"
#  "Launcher (gnome-terminal) added to bottom panel" etc.
#

class PanelAddedChange (PanelChange):
    def __init__ (self, source, delegate, id):
        PanelChange.__init__ (self, source, delegate, id)
    def get_short_description (self):
        return _("Panel '%s' added") % self.id
    
class PanelRemovedChange (PanelChange):
    def __init__ (self, source, delegate, id):
        PanelChange.__init__ (self, source, delegate, id)
    def get_short_description (self):
        return _("Panel '%s' removed") % self.id

class PanelAppletAddedChange (PanelChange):
    def __init__ (self, source, delegate, id):
        PanelChange.__init__ (self, source, delegate, id)
    def get_short_description (self):
        return _("Panel applet '%s' added") % self.id

class PanelAppletRemovedChange (PanelChange):
    def __init__ (self, source, delegate, id):
        PanelChange.__init__ (self, source, delegate, id)
    def get_short_description (self):
        return _("Panel applet '%s' removed") % self.id

class PanelObjectAddedChange (PanelChange):
    def __init__ (self, source, delegate, id):
        PanelChange.__init__ (self, source, delegate, id)
    def get_short_description (self):
        return _("Panel object '%s' added") % self.id
    
class PanelObjectRemovedChange (PanelChange):
    def __init__ (self, source, delegate, id):
        PanelChange.__init__ (self, source, delegate, id)
    def get_short_description (self):
        return _("Panel object '%s' removed") % self.id

class PanelDelegate (userprofile.SourceDelegate):
    class PanelThing:
        def __init__ (self, id, added, removed):
            self.id      = id
            self.added   = added
            self.removed = removed
    
    class PanelToplevel (PanelThing):
        def __init__ (self, id, added = False, removed = False):
            PanelDelegate.PanelThing.__init__ (self, id, added, removed)
            
            # FIXME: which attributes do we really need?
            # self.name        = self.client.get_string (PANEL_KEY_BASE + "/toplevels/" + toplevel_id + "/name")
            # self.orientation = self.client.get_string (PANEL_KEY_BASE + "/toplevels/" + toplevel_id + "/orientation")
            # self.expand      = self.client.get_bool   (PANEL_KEY_BASE + "/toplevels/" + toplevel_id + "/expand")
        
    class PanelApplet (PanelThing):
        def __init__ (self, id, added = False, removed = False):
            PanelDelegate.PanelThing.__init__ (self, id, added, removed)
            
            # FIXME: which attributes do we really need?
            # self.toplevel_id = self.client.get_string (PANEL_KEY_BASE + "/applets/" + applet_id + "/toplevel_id")
            # self.bonobo_iid  = self.client.get_string (PANEL_KEY_BASE + "/applets/" + applet_id + "/bonobo_iid")

    class PanelObject (PanelThing):
        def __init__ (self, id, added = False, removed = False):
            PanelDelegate.PanelThing.__init__ (self, id, added, removed)
            
            # FIXME: which attributes do we really need?
            # self.toplevel_id = self.client.get_string (PANEL_KEY_BASE + "/objects/" + object_id + "/toplevel_id")
            # self.object_type = self.client.get_string (PANEL_KEY_BASE + "/objects/" + object_id + "/object_type")

    def __init__ (self, source):
        userprofile.SourceDelegate.__init__ (self, _("Panel"), source, PANEL_KEY_BASE)
        self.client = gconf.client_get_default ()

        self.toplevels = {}
        self.applets = {}
        self.objects = {}
        self.__read_panel_config ()

    def __read_panel_config (self):
        for id in self.client.get_list (PANEL_KEY_BASE + "/general/toplevel_id_list", gconf.VALUE_STRING):
            if not self.toplevels.has_key (id):
                self.toplevels[id] = PanelDelegate.PanelToplevel (id, False)
        for id in self.client.get_list (PANEL_KEY_BASE + "/general/applet_id_list", gconf.VALUE_STRING):
            if not self.applets.has_key (id):
                self.applets[id] = PanelDelegate.PanelApplet (id, False)
        for id in self.client.get_list (PANEL_KEY_BASE + "/general/object_id_list", gconf.VALUE_STRING):
            if not self.objects.has_key (id):
                self.objects[id] = PanelDelegate.PanelObject (id, False)

    def __handle_id_list_change (self, change, dict, thing_class, added_class, removed_class):
        if not change.value or \
               change.value.type != gconf.VALUE_LIST or \
               change.value.get_list_type () != gconf.VALUE_STRING:
                return True

        id_list = []
        for v in change.value.get_list ():
            id_list.append (v.get_string ())
            
        added = []
        for id in id_list:
            if dict.has_key (id) and not dict[id].removed:
                continue
            if not dict.has_key (id):
                dict[id] = thing_class (id, True)
            else:
                dict[id].added   = True
                dict[id].removed = False
            added.append (id)

        removed = []
        for id in dict:
            if id in id_list:
                continue
            if dict.has_key (id) and not dict[id].removed:
                dict[id].removed = True
                removed.append (id)

        for id in added:
            self.source.emit ("changed", added_class (self.source, self, id))
        for id in removed:
            self.source.emit ("changed", removed_class (self.source, self, id))

        return True

    def handle_change (self, change):
        if change.key.startswith (PANEL_KEY_BASE + "/toplevels/"):
            toplevel_id = change.key.split ("/")[4]
            if not self.toplevels.has_key (toplevel_id) or \
               self.toplevels[toplevel_id].added or \
               self.toplevels[toplevel_id].removed:
                return True
        
        elif change.key.startswith (PANEL_KEY_BASE + "/objects/"):
            object_id = change.key.split ("/")[4]
            if not self.objects.has_key (object_id) or \
               self.objects[object_id].added or \
               self.objects[object_id].removed:
                return True
        
        elif change.key.startswith (PANEL_KEY_BASE + "/applets"):
            applet_id = change.key.split ("/")[4]
            if not self.applets.has_key (applet_id) or \
               self.applets[applet_id].added or \
               self.applets[applet_id].removed:
                return True
        
        elif change.key == PANEL_KEY_BASE + "/general/toplevel_id_list":
            return self.__handle_id_list_change (change,
                                                 self.toplevels,
                                                 PanelDelegate.PanelToplevel,
                                                 PanelAddedChange,
                                                 PanelRemovedChange)
            
        elif change.key == PANEL_KEY_BASE + "/general/applet_id_list":
            return self.__handle_id_list_change (change,
                                                 self.applets,
                                                 PanelDelegate.PanelApplet,
                                                 PanelAppletAddedChange,
                                                 PanelAppletRemovedChange)
            
        elif change.key == PANEL_KEY_BASE + "/general/object_id_list":
            return self.__handle_id_list_change (change,
                                                 self.objects,
                                                 PanelDelegate.PanelObject,
                                                 PanelObjectAddedChange,
                                                 PanelObjectRemovedChange)

        return False

    def __copy_dir (self, src_client, dst_client, dst_address, dir):
        for entry in src_client.all_entries (dir):
            if entry.get_schema_name ():
                gconfsource.associate_schema (dst_address, entry.key, entry.get_schema_name ())
            if entry.value and not entry.get_is_default ():
                dst_client.set (entry.key, entry.value)
        for subdir in src_client.all_dirs (dir):
            self.__copy_dir (src_client, dst_client, dst_address, subdir)

    def __get_current_list (self, dict):
        id_list = []
        for id in dict:
            if dict[id].added:
                continue
            id_list.append (id)
        return id_list

    def __commit_added_change (self, change, mandatory, dict, id_list_name, dir_name):
        if not dict.has_key (change.id):
            return
        
        thing = dict[change.id]
        if not thing.added:
            return

        (client, address) = self.source.get_committing_client_and_address (mandatory)

        self.__copy_dir (self.client, client, address, PANEL_KEY_BASE + "/" + dir_name + "/" + thing.id)
        
        id_list = self.__get_current_list (dict)
        id_list.append (thing.id)
        client.set_list (PANEL_KEY_BASE + "/general/" + id_list_name, gconf.VALUE_STRING, id_list)
        
        thing.added = False
        
    def __commit_removed_change (self, change, mandatory, dict, id_list_name, dir_name):
        if not dict.has_key (change.id):
            return

        thing = dict[change.id]
        if not thing.removed:
            return

        (client, address) = self.source.get_committing_client_and_address (mandatory)

        id_list = self.__get_current_list (dict)
        if thing.id in id_list:
            id_list.remove (thing.id)
            client.set_list (PANEL_KEY_BASE + "/general/" + id_list_name, gconf.VALUE_STRING, id_list)
            gconfsource.recursive_unset (client, PANEL_KEY_BASE + "/" + dir_name + "/" + thing.id)
        
        del dict[change.id]

    def commit_change (self, change, mandatory = False):
        if isinstance (change, PanelAddedChange):
            self.__commit_added_change (change,
                                        mandatory,
                                        self.toplevels,
                                        "toplevel_id_list",
                                        "toplevels")
        elif isinstance (change, PanelRemovedChange):
            self.__commit_removed_change (change,
                                          mandatory,
                                          self.toplevels,
                                          "toplevel_id_list",
                                          "toplevels")
        elif isinstance (change, PanelAppletAddedChange):
            self.__commit_added_change (change,
                                        mandatory,
                                        self.applets,
                                        "applet_id_list",
                                        "applets")
        elif isinstance (change, PanelAppletRemovedChange):
            self.__commit_removed_change (change,
                                          mandatory,
                                          self.applets,
                                          "applet_id_list",
                                          "applets")
        elif isinstance (change, PanelObjectAddedChange):
            self.__commit_added_change (change,
                                        mandatory,
                                        self.objects,
                                        "object_id_list",
                                        "objects")
        elif isinstance (change, PanelObjectRemovedChange):
            self.__commit_removed_change (change,
                                          mandatory,
                                          self.objects,
                                          "object_id_list",
                                          "objects")

    def start_monitoring (self):
        """Start monitoring for configuration changes."""
        # Nothing to do here
        pass

    def stop_monitoring (self):
        """Stop monitoring for configuration changes."""
        # Nothing to do here
        pass

    def sync_changes (self):
        # Nothing to do here
        pass
    
    def apply (self):
        # Nothing to do here
        pass

def get_gconf_delegate (source):
    return PanelDelegate (source)

#
# Unit tests
#
def run_unit_tests ():
    import gobject
    import os
    import os.path
    import time
    import tempfile
    import shutil

    # Clear out any cruft
    os.system ("gconftool-2 --recursive-unset %s/toplevels/foo" % PANEL_KEY_BASE)
    os.system ("gconftool-2 --recursive-unset %s/objects/foo" % PANEL_KEY_BASE)
    os.system ("gconftool-2 --recursive-unset %s/applets/foo" % PANEL_KEY_BASE)
    time.sleep (1)

    client = gconf.client_get_default ()
    
    for id_list_name in ("toplevel_id_list", "object_id_list", "applet_id_list"):
        id_list = client.get_list (PANEL_KEY_BASE + "/general/" + id_list_name, gconf.VALUE_STRING)
        while "foo" in id_list:
            id_list.remove ("foo")
        client.set_list (PANEL_KEY_BASE + "/general/" + id_list_name, gconf.VALUE_STRING, id_list)

    temp_path = tempfile.mkdtemp (prefix = "test-paneldelegate-")

    # Create a dummy source with a PanelDelegate
    class TempSource (userprofile.ProfileSource):
        def __init__ (self, temp_path):
            userprofile.ProfileSource.__init__ (self, "panel-temp")
            self.delegates.append (PanelDelegate (self))
            self.temp_path = temp_path
            self.defaults_client = None
            self.defaults_address = None
            self.mandatory_client = None
            self.mandatory_address = None
        def get_committing_client_and_address (self, mandatory):
            if not mandatory:
                if not self.defaults_client:
                    (client, address) = gconfsource.get_client_and_address_for_path (
                        os.path.join (self.temp_path, ".gconf.xml.defaults"))
                    self.defaults_client = client
                    self.defaults_address = address
                return (self.defaults_client, self.defaults_address)
            else:
                if not self.mandatory_client:
                    (client, address) = gconfsource.get_client_and_address_for_path (
                        os.path.join (self.temp_path, ".gconf.xml.mandatory"))
                    self.mandatory_client = client
                    self.mandatory_address = address
                return (self.mandatory_client, self.mandatory_address)
    gobject.type_register (TempSource)
    source = TempSource (temp_path)

    # Set up the client
    def handle_notify (client, cnx_id, entry, source):
        source.emit_change (gconfsource.GConfChange (source, entry.key, entry.value))
    client.add_dir (PANEL_KEY_BASE + "", gconf.CLIENT_PRELOAD_RECURSIVE)
    notify_id = client.notify_add (PANEL_KEY_BASE + "", handle_notify, source)

    # Trap filtered changes
    global changes
    changes = []
    def handle_change (source, change):
        global changes
        changes.append (change)
    source.connect ("changed", handle_change)

    # Need to run the mainloop to get notifications.
    # The notification is only dispatched once the set
    # operation has complete
    main_loop = gobject.MainLoop ()
    def poll (main_loop):
        while main_loop.get_context ().pending ():
            main_loop.get_context ().iteration (False)

    # Recursively copy a dir
    def copy_dir (client, dst, src):
        for entry in client.all_entries (src):
            key = dst + "/" + os.path.basename (entry.key)
            client.set (key, entry.value)
        for dir in client.all_dirs (src):
            subdir = os.path.basename (dir)
            copy_dir (client, dst + "/" + subdir, src + "/" + subdir)

    # Set random uninterpreted key
    show_program_list = client.get_bool (PANEL_KEY_BASE + "/general/show_program_list")
    client.set_bool (PANEL_KEY_BASE + "/general/show_program_list", not show_program_list)
    poll (main_loop)

    # Set up a panel on the right
    copy_dir (client, PANEL_KEY_BASE + "/toplevels/foo", PANEL_KEY_BASE + "/toplevels/top_panel")
    poll (main_loop)

    # Set up a clock applet
    copy_dir (client, PANEL_KEY_BASE + "/applets/foo", PANEL_KEY_BASE + "/applets/clock")
    client.set_string (PANEL_KEY_BASE + "/applets/foo/toplevel_id", "foo")
    poll (main_loop)
    
    # Set up a menu bar
    copy_dir (client, PANEL_KEY_BASE + "/objects/foo", PANEL_KEY_BASE + "/objects/menu_bar")
    client.set_string (PANEL_KEY_BASE + "/objects/foo/toplevel_id", "foo")
    poll (main_loop)

    # Add the new panel to the list of panels
    toplevels = client.get_list (PANEL_KEY_BASE + "/general/toplevel_id_list", gconf.VALUE_STRING)
    toplevels.append ("foo")
    client.set_list (PANEL_KEY_BASE + "/general/toplevel_id_list", gconf.VALUE_STRING, toplevels)
    poll (main_loop)
    
    # Add the new clock to the list of applets
    applets = client.get_list (PANEL_KEY_BASE + "/general/applet_id_list", gconf.VALUE_STRING)
    applets.append ("foo")
    client.set_list (PANEL_KEY_BASE + "/general/applet_id_list", gconf.VALUE_STRING, applets)
    poll (main_loop)

    # Add the new menu bar to the list of objects
    objects = client.get_list (PANEL_KEY_BASE + "/general/object_id_list", gconf.VALUE_STRING)
    objects.append ("foo")
    client.set_list (PANEL_KEY_BASE + "/general/object_id_list", gconf.VALUE_STRING, objects)
    poll (main_loop)

    time.sleep (3)

    client.set_int (PANEL_KEY_BASE + "/toplevels/foo/hide_delay", 5)
    poll (main_loop)

    assert len (changes) == 4
    for change in changes:
        assert isinstance (change, userprofile.ProfileChange)
    assert isinstance (changes[0], gconfsource.GConfChange)
    assert changes[0].key == PANEL_KEY_BASE + "/general/show_program_list"
    for change in changes[1:4]:
        assert isinstance (change, PanelChange)
        assert change.id == "foo"

    (
        PANEL_ADDED,
        PANEL_APPLET_ADDED,
        PANEL_OBJECT_ADDED
    ) = range (1, 4)
    
    assert isinstance (changes[PANEL_ADDED], PanelAddedChange)
    assert isinstance (changes[PANEL_APPLET_ADDED], PanelAppletAddedChange)
    assert isinstance (changes[PANEL_OBJECT_ADDED], PanelObjectAddedChange)

    source.commit_change (changes[PANEL_ADDED], False)
    
    os.system ("gconftool-2 --shutdown")
    time.sleep (1)

    assert os.access (os.path.join (temp_path, ".gconf.xml.defaults/apps/panel/general/%gconf.xml"), os.F_OK)
    assert os.access (os.path.join (temp_path, ".gconf.xml.defaults/apps/panel/toplevels/foo/%gconf.xml"), os.F_OK)

    changes = []

    # Remove the menu bar again
    while "foo" in objects:
        objects.remove ("foo")
    client.set_list (PANEL_KEY_BASE + "/general/object_id_list", gconf.VALUE_STRING, objects)
    poll (main_loop)
    
    # Remove the clock again
    while "foo" in applets:
        applets.remove ("foo")
    client.set_list (PANEL_KEY_BASE + "/general/applet_id_list", gconf.VALUE_STRING, applets)
    poll (main_loop)
    
    # Remove the panel again
    while "foo" in toplevels:
        toplevels.remove ("foo")
    client.set_list (PANEL_KEY_BASE + "/general/toplevel_id_list", gconf.VALUE_STRING, toplevels)
    poll (main_loop)

    # Set random uninterpreted key back again
    client.set_bool (PANEL_KEY_BASE + "/general/show_program_list", show_program_list)
    poll (main_loop)
    
    # Shutdown the client again
    client.notify_remove (notify_id)
    client.remove_dir (PANEL_KEY_BASE + "")

    assert len (changes) == 4
    for change in changes:
        assert isinstance (change, userprofile.ProfileChange)
    for change in changes[:3]:
        assert isinstance (change, PanelChange)
        assert change.id == "foo"
    assert isinstance (changes[3], gconfsource.GConfChange)
    assert changes[3].key == PANEL_KEY_BASE + "/general/show_program_list"

    (
        PANEL_OBJECT_REMOVED,
        PANEL_APPLET_REMOVED,
        PANEL_REMOVED
    ) = range (3)
    
    assert isinstance (changes[PANEL_OBJECT_REMOVED], PanelObjectRemovedChange)
    assert isinstance (changes[PANEL_APPLET_REMOVED], PanelAppletRemovedChange)
    assert isinstance (changes[PANEL_REMOVED], PanelRemovedChange)

    source.commit_change (changes[PANEL_REMOVED], False)

    dprint ("Committed changes to %s", temp_path)
    os.system ("gconftool-2 --shutdown")
    time.sleep (1)
    
    assert not os.access (os.path.join (temp_path, ".gconf.xml.defaults/apps/panel/toplevels/foo/%gconf.xml"), os.F_OK)

    # Bye, bye cruft
    os.system ("gconftool-2 --recursive-unset %s/toplevels/foo" % PANEL_KEY_BASE)
    os.system ("gconftool-2 --recursive-unset %s/objects/foo" % PANEL_KEY_BASE)
    os.system ("gconftool-2 --recursive-unset %s/applets/foo" % PANEL_KEY_BASE)

    shutil.rmtree (temp_path, True)



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
# Copyright (C) 2004 GNOME Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

#
# All the text was copied from gedit-close-confirmation-dialog.c
#

import gtk
import gettext

class SaveConfirmationAlert (gtk.MessageDialog):
    def __init__ (self, parent_window, profile_name, seconds):
        gtk.MessageDialog.__init__ (self, parent_window, 0, gtk.MESSAGE_WARNING, gtk.BUTTONS_NONE)

        self.set_destroy_with_parent (True)
        self.set_title ("")

        self.add_button (_("Close _Without Saving"), gtk.RESPONSE_NO)
        self.add_button (gtk.STOCK_CANCEL,           gtk.RESPONSE_CANCEL)
        self.add_button (gtk.STOCK_SAVE,             gtk.RESPONSE_YES)

        self.set_default_response (gtk.RESPONSE_YES)

        self.set_markup ("<b>" +
                         _("Save changes to profile \"%s\" before closing?") % profile_name
                         + "</b>")

        if seconds < 55:
            secondary_msg = gettext.ngettext ("If you don't save, changes from the last %ld second "
                                              "will be permanently lost.",
                                              "If you don't save, changes from the last %ld seconds "
                                              "will be permanently lost.",
                                              seconds) % seconds
        elif seconds < 75:
            secondary_msg = _("If you don't save, changes from the last minute will be permanently lost.")
            
        elif seconds < 110:
            seconds -= 60
            secondary_msg = gettext.ngettext ("If you don't save, changes from the last minute and %ld "
                                              "second will be permanently lost.",
                                              "If you don't save, changes from the last minute and %ld "
                                              "seconds will be permanently lost.",
                                              seconds) % seconds
        elif seconds < 3600:
            minutes = seconds / 60
            secondary_msg = gettext.ngettext ("If you don't save, changes from the last %ld minute "
                                              "will be permanently lost.",
                                              "If you don't save, changes from the last %ld minutes "
                                              "will be permanently lost.",
                                              minutes) % minutes
        elif seconds < 7200:
            seconds -= 3600;
            minutes = seconds / 60;
            if minutes < 5:
                secondary_msg = _("If you don't save, changes from the last hour "
                                  "will be permanently lost.")
            else:
                secondary_msg = gettext.ngettext ("If you don't save, changes from the last hour and %d "
                                                  "minute will be permanently lost.",
                                                  "If you don't save, changes from the last hour and %d "
                                                  "minutes will be permanently lost.",
                                                  minutes) % minutes
        else:
            hours = seconds / 3600;
            secondary_msg = gettext.ngettext ("If you don't save, changes from the last hour "
                                              "will be permanently lost.",
                                              "If you don't save, changes from the last %d hours "
                                              "will be permanently lost.",
                                              hours) % hours
                                                            
        self.format_secondary_text (secondary_msg)



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import os
import os.path
import sys
import pwd
import gettext
import locale
from config import *

(
    DEBUG_USERPROFILE,
    DEBUG_STORAGE,
    DEBUG_PROTOSESSION,
    DEBUG_USERMOD,
    DEBUG_DIRMONITOR,
    DEBUG_GCONFSOURCE,
    DEBUG_PANELDELEGATE,
    DEBUG_FILESSOURCE,
    DEBUG_MOZILLASOURCE,
    DEBUG_ADMINTOOL,
    DEBUG_USERDB,
) = range (11)

debug_modules = {
    DEBUG_USERPROFILE   : ("user-profile",   False),
    DEBUG_STORAGE       : ("storage",        False),
    DEBUG_PROTOSESSION  : ("proto-session",  False),
    DEBUG_USERMOD       : ("usermod",        False),
    DEBUG_DIRMONITOR    : ("dir-monitor",    False),
    DEBUG_GCONFSOURCE   : ("gconf-source",   False),
    DEBUG_PANELDELEGATE : ("panel-delegate", False),
    DEBUG_FILESSOURCE   : ("files-source",   False),
    DEBUG_MOZILLASOURCE : ("mozilla-source", False),
    DEBUG_ADMINTOOL     : ("admin-tool",     False),
    DEBUG_USERDB        : ("user-db",        False),
}

def init_debug_modules ():
    debug_value = os.getenv ("SABAYON_DEBUG")
    if not debug_value:
        return

    if debug_value == "help":
        print "Valid options for the SABAYON_DEBUG environment variable are:\n"
        print "    all"
        for module in debug_modules:
            print "    %s" % debug_modules[module][0]
        print "You may supply a list of modules separated by a colon (:)"
        print "You may also supply an optional hex debug mask to a module, e.g. foo=0xF8"
        sys.exit (1)
    elif debug_value == "all":
        for module in debug_modules:
            debug_modules[module] = (debug_modules[module][0], True)
    else:
        for item in debug_value.split (":"):
            item = item.split("=")
            key = item[0]
            if len(item) > 1:
                value = int(item[1],16)
            else:
                value = ~0
            for module in debug_modules:
                if debug_modules[module][0] == key:
                    debug_modules[module] = (key, value)
                    break

init_debug_modules ()

def debug_print (module, message, mask=~0):
    assert debug_modules.has_key(module)
    if not debug_modules[module][1] & mask:
        return
    print "(%d) %s: %s" % (os.getpid (), debug_modules[module][0], message)

class GeneralError (Exception):
    def __init__ (self, msg):
        Exception.__init__ (self, msg)

unit_tests_homedir = None
def set_home_dir_for_unit_tests (homedir):
    global unit_tests_homedir
    unit_tests_homedir = homedir

def get_home_dir ():
    if unit_tests_homedir:
        return unit_tests_homedir
    try:
        pw = pwd.getpwuid (os.getuid ())
        if pw.pw_dir != "":
            return pw.pw_dir
    except KeyError:
        pass
    
    if os.environ.has_key ("HOME"):
        return os.environ["HOME"]
    else:
        raise GeneralError (_("Cannot find home directory: not set in /etc/passwd and no value for $HOME in environment"))

def get_user_name ():
    try:
        pw = pwd.getpwuid (os.getuid ())
        if pw.pw_name != "":
            return pw.pw_name
    except KeyError:
        pass
    
    if os.environ.has_key ("USER"):
        return os.environ["USER"]
    else:
        raise GeneralError (_("Cannot find username: not set in /etc/passwd and no value for $USER in environment"))

def print_exception ():
    import traceback
    import sys
    traceback.print_exc(file=sys.stderr)

def init_gettext ():
    """Binds _() to gettext.gettext() in the global namespace. Run
    util.init_gettext() at the entry point to any script and you'll
    be able to use _() to mark strings for translation.
    """
    locale.setlocale (locale.LC_ALL, "")
    gettext.install (PACKAGE, os.path.join (DATADIR, "locale"))

def run_unit_tests ():
    home_dir = get_home_dir ()
    assert home_dir != ""
    assert get_user_name () != ""
    set_home_dir_for_unit_tests ("foo")
    assert get_home_dir () == "foo"
    set_home_dir_for_unit_tests (None)
    assert get_home_dir () == home_dir



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import gtk
import gconf

class GConfModel (gtk.TreeStore):
    (
        COLUMN_NAME,
        COLUMN_TYPE,
        COLUMN_VALUE
    ) = range (3)
        
    def __init__ (self, client):
        gtk.TreeStore.__init__ (self, str, str, str)

        self.client = client
        self.__append_dir_contents (None, "/")

    def __append_dir_contents (self, parent_iter, dir):
        subdirs = list (self.client.all_dirs (dir))
        subdirs.sort ()
        for subdir in subdirs:
            iter = self.append (parent_iter)
            self.set (iter,
                      self.COLUMN_NAME, subdir[subdir.rfind ("/") + 1:])
            self.__append_dir_contents (iter, subdir)

        entries = list (self.client.all_entries (dir))
        entries.sort (lambda a, b: cmp (a.key, b.key))
        for entry in entries:
            iter = self.append (parent_iter)
            self.set (iter,
                      self.COLUMN_NAME, entry.key[entry.key.rfind ("/") + 1:])
            if entry.value:
                self.set (iter,
                          self.COLUMN_TYPE,  self.__type_to_string (entry.value.type),
                          self.COLUMN_VALUE, entry.value.to_string ())
            else:
                self.set (iter,
                          self.COLUMN_TYPE,  _("<no type>"),
                          self.COLUMN_VALUE, _("<no value>"))

    def __type_to_string (self, type):
        if type == gconf.VALUE_STRING:
            return _("string")
        elif type == gconf.VALUE_INT:
            return _("integer")
        elif type == gconf.VALUE_FLOAT:
            return _("float")
        elif type == gconf.VALUE_BOOL:
            return _("boolean")
        elif type == gconf.VALUE_SCHEMA:
            return _("schema")
        elif type == gconf.VALUE_LIST:
            return _("list")
        elif type == gconf.VALUE_PAIR:
            return _("pair")
        else:
            return _("<no type>")

class GConfViewer (gtk.Window):
    def __init__ (self, xml_backend_dir, parent_window):
        gtk.Window.__init__ (self, gtk.WINDOW_TOPLEVEL)
        
        self.backend_address = "xml:readonly:" + xml_backend_dir
        self.engine = gconf.engine_get_for_address ("xml:readonly:" + xml_backend_dir)
        self.client = gconf.client_get_for_engine (self.engine)

        self.set_title (_("All Your Settings Are Belong To Us"))
        self.set_icon_name ("sabayon")
        self.set_transient_for (parent_window)
        self.set_destroy_with_parent (True)
        self.set_default_size (480, 380)

        self.scrolled = gtk.ScrolledWindow ()
        self.scrolled.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scrolled.set_shadow_type (gtk.SHADOW_IN)
        self.scrolled.show ()
        self.add (self.scrolled)

        self.__setup_treeview ()
    
    def __setup_treeview (self):
        self.gconf_model = GConfModel (self.client)

        self.treeview = gtk.TreeView (self.gconf_model)
        self.treeview.show ()
        self.scrolled.add (self.treeview)

        self.treeview.get_selection ().set_mode (gtk.SELECTION_NONE)
        self.treeview.set_headers_visible (True)

        c = gtk.TreeViewColumn (_("Name"),
                                gtk.CellRendererText (),
                                text = GConfModel.COLUMN_NAME)
        self.treeview.append_column (c)
        
        c = gtk.TreeViewColumn (_("Type"),
                                gtk.CellRendererText (),
                                text = GConfModel.COLUMN_TYPE)
        self.treeview.append_column (c)
        
        c = gtk.TreeViewColumn (_("Value"),
                                gtk.CellRendererText (),
                                text = GConfModel.COLUMN_VALUE)
        self.treeview.append_column (c)

        self.treeview.expand_all ()



#!/usr/bin/env python
import sys
import string
import pwd
import os
import libxml2
import config
import util

defaultConf="""<profiles>
  <default profile="default"/>
</profiles>"""

def libxml2_no_error_callback(ctx, str):
    pass

libxml2.registerErrorHandler(libxml2_no_error_callback, "")

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_USERDB, fmt % args)

class UserDatabaseException (Exception):
    pass

class UserDatabase:
    """An encapsulation of the database which maintains an
    association between users and profiles.

    This database is stored by default in
    $(sysconfdir)/desktop-profiles/users.xml and contains a
    list of users and the profile associated with each of
    those users.

    A profile can be reference by either its name (in which case
    the profile is stored at /etc/desktop-profiles/$(name).zip),
    an absolute path or a http/file URL.
    """
    def __init__ (self, db_file = None):
        """Create a UserDatabase object.

        @db_file: an (optional) path which specifes the location
        of the database file. If not specified, the default
        location of /etc/desktop-profiles/users.xml is used.
        """
        if db_file is None:
            file = os.path.join (config.PROFILESDIR, "users.xml")
	elif db_file[0] != '/':
            file = os.path.join (config.PROFILESDIR, db_file)
	else:
	    file = db_file
	self.file = file;
	self.modified = 0
	dprint("New UserDatabase(%s) object\n" % self.file)

	try:
	    self.doc = libxml2.readFile(file, None, libxml2.XML_PARSE_NOBLANKS)
	except:
	    # TODO add fallback to last good database
	    dprint("failed to parse %s falling back to default conf\n" %
	           self.file)
	    self.doc = None
	if self.doc == None:
	    self.doc = libxml2.readMemory(defaultConf, len(defaultConf),
	                                  None, None,
	                                  libxml2.XML_PARSE_NOBLANKS);

    def __del__ (self):
        if self.doc != None:
	    self.doc.freeDoc()

    def get_profile (self, username):
        """Look up the profile for a given username.

        @username: the user whose profile location should be
        returned.

        Return value: the location of the profile. The location
        should  be in a suitable form for constructing a
        ProfileStorage object.
        """
	try:
	    query = "/profiles/user[@name='%s']" % username
	    user = self.doc.xpathEval(query)[0]
	    profile = user.prop("profile")
	except:
	    profile = None
	if profile is None or profile == "":
	    try:
	        query = "string(/profiles/default[1]/@profile)"
	        profile = self.doc.xpathEval(query)
	    except:
	        profile = None
	if profile == "":
	    profile = None
	if profile is None:
	    return None
	try:
	    uri = libxml2.parseURI(profile);
	    if uri.scheme() is None:
	        # it is a file path
		if profile[0] != '/':
		    profile = os.path.join (config.PROFILESDIR, profile)
		if profile[-4:] != ".zip":
		    profile = profile + ".zip"
	    else:
	        # TODO need to make a local copy or use the local copy
		profile = profile

	except:
	    # we really expect an URI there
	    return None

        # TODO Check the resulting file path exists
        return profile

    def __save_as(self, filename = None):
        """Save the current version to the given filename"""
	if filename == None:
	    filename = self.file

	dprint("Saving UserDatabase to %s\n", filename)
	try:
	    os.rename(filename, filename + ".bak")
	    backup = 1
	except:
	    backup = 0
	    pass

	try:
	    f = open(filename, 'w')
	except:
	    if backup == 1:
	        try:
		    os.rename(filename + ".bak", filename)
		    dprint("Restore from %s.bak\n", filename)
		except:
		    dprint("Failed to restore from %s.bak\n", filename)

	    raise UserDatabaseException(
	              _("Could not open %s for writing") % filename)
	try:
	    f.write(self.doc.serialize("UTF-8", format=1))
	    f.close()
	except:
	    if backup == 1:
	        try:
		    os.rename(filename + ".bak", filename)
		    dprint("Restore from %s.bak\n", filename)
		except:
		    dprint("Failed to restore from %s.bak\n", filename)

	    raise UserDatabaseException(
	              _("Failed to save UserDatabase to %s") % filename)
	
	self.modified = 0

    def set_profile (self, username, profile):
        """Set the profile for a given username.

        @username: the user whose profile location should be
        set.
        @profile: the location of the profile.
        """
	self.modified = 0
	try:
	    query = "/profiles/user[@name='%s']" % username
	    user = self.doc.xpathEval(query)[0]
	    oldprofile = user.prop("profile")
	    if oldprofile != profile:
	        user.setProp("profile", profile)
		self.modified = 1
	except:
	    try:
		profiles = self.doc.xpathEval("/profiles")[0]
	    except:
		raise UserDatabaseException(
			  _("File %s is not a profile configuration") %
                                           (self.file))
	    try:
		user = profiles.newChild(None, "user", None)
		user.setProp("name", username)
		user.setProp("profile", profile)
	    except:
		raise UserDatabaseException(
			  _("Failed to add user %s to profile configuration") %
                                           (username))
	    self.modified = 1
	if self.modified == 1:
	    self.__save_as()
	
    def get_profiles (self):
        """Return the list of currently available profiles.
        This is basically just list of zip files in
        /etc/desktop-profiles, each without the .zip extension.
        """
	list = []
	try:
	    for file in os.listdir('/etc/desktop-profiles'):
	        if file[-4:] != ".zip":
		    continue
		list.append(file[0:-4])
	except:
	    dprint("Failed to read directory(%s)\n" % (config.PROFILESDIR))
	# TODO: also list remote profiles as found in self.doc
	return list

    def get_users (self):
        """Return the list of users on the system. These should
        be real users - i.e. should not include system users
        like nobody, gdm, nfsnobody etc.
        """
	list = []
	try:
	    users = pwd.getpwall()
	except:
	    raise UserDatabaseException(_("Failed to get the user list"))

	for user in pwd.getpwall():
	    try:
	        # remove non-users
		if user[2] < 1000:
		    continue
		if user[0] in list:
		    continue
		if user[6] == "" or string.find(user[6], "nologin") != -1:
		    continue
		list.append(user[0])
	    except:
		pass
	return list



user_database = None
def get_database ():
    """Return a UserDatabase singleton"""
    global user_database
    if user_database is None:
        user_database = UserDatabase ()
    return user_database

#
# Unit tests
#
def run_unit_tests ():
    testfile = "/tmp/test_users.xml"
    try:
        os.unlink(testfile)
    except:
        pass
    db = UserDatabase(testfile)
    res = db.get_profile("localuser")
    if res is None:
        print "get_profile failed to return a value"
    if res[-29:] != "/desktop-profiles/default.zip":
        print "get_profile returned the wrong value"
    db.set_profile("localuser", "groupA")
    res = db.get_profile("localuser")
    if res is None:
        print "get_profile failed to return a value"
    if res[-28:] != "/desktop-profiles/groupA.zip":
        print "get_profile returned a wrong value, expected groupA.zip got", res
    db.set_profile("localuser", "groupB")
    res = db.get_profile("localuser")
    if res is None:
        print "get_profile failed to return a value"
    if res[-28:] != "/desktop-profiles/groupB.zip":
        print "get_profile returned a wrong value, expected groupB.zip got", res
        

if __name__ == "__main__":
    util.init_gettext ()
    run_unit_tests()



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import os
import os.path
import errno
import shutil
import tempfile
import pwd
import gobject
import gtk
import gtk.glade
import storage
import protosession
import editorwindow
import util
from config import *

def dprint (fmt, *args):
    util.debug_print (util.DEBUG_ADMINTOOL, fmt % args)

def _get_profile_path_for_name (profile_name):
    return os.path.join (PROFILESDIR, profile_name + ".zip")

class ProfilesModel (gtk.ListStore):
    (
        COLUMN_NAME,
    ) = range (1)

    def __init__ (self):
        gtk.ListStore.__init__ (self, str)
        self.reload ()

    def reload (self):
        self.clear ()
        profile_files = []
        try:
            profile_files = os.listdir (PROFILESDIR)
        except OSError, err:
            if err.errno != errno.ENOENT:
                raise err
                                        
        for file in profile_files:
            if not file.endswith (".zip"):
                continue
                
            row = self.append ()
            self.set (row, self.COLUMN_NAME, file[:-len (".zip")])

class NewProfileDialog:
    def __init__ (self, profiles_model):
        self.profiles_model = profiles_model
        
        glade_file = os.path.join (GLADEDIR, "sabayon.glade")
        self.xml = gtk.glade.XML (glade_file, "new_profile_dialog")
        
        self.dialog = self.xml.get_widget ("new_profile_dialog")
        self.dialog.connect ("destroy", gtk.main_quit)
        self.dialog.set_default_response (gtk.RESPONSE_ACCEPT)
        self.dialog.set_icon_name ("sabayon")

        self.create_button = self.xml.get_widget ("new_profile_create_button")
        self.create_button.set_sensitive (False)

        self.name_entry = self.xml.get_widget ("new_profile_name_entry")
        self.name_entry.connect ("changed", self.__name_entry_changed)
        self.name_entry.set_activates_default (True)
        
        self.base_combo = self.xml.get_widget ("new_profile_base_combo")
        self.base_combo.set_model (self.profiles_model)

        renderer = gtk.CellRendererText ()
        self.base_combo.pack_start (renderer, True)
        self.base_combo.set_attributes (renderer, text = ProfilesModel.COLUMN_NAME)

    def __name_entry_changed (self, entry):
        text = entry.get_text ()
        if not text or text.isspace ():
            self.create_button.set_sensitive (False)
        else:
            self.create_button.set_sensitive (True)

    def run (self, parent):
        self.name_entry.grab_focus ()
        self.dialog.set_transient_for (parent)
        self.dialog.present ()
        response = self.dialog.run ()
        self.dialog.hide ()
        
        if response != gtk.RESPONSE_ACCEPT:
            return (None, None)

        iter = self.base_combo.get_active_iter ()
        if iter:
            base = self.profiles_model.get_value (iter, ProfilesModel.COLUMN_NAME)
        else:
            base = None
        
        return (self.name_entry.get_text (), base)

class ProfilesDialog:
    def __init__ (self):
        assert os.geteuid () == 0
        
        glade_file = os.path.join (GLADEDIR, "sabayon.glade")
        self.xml = gtk.glade.XML (glade_file, "profiles_dialog")

        self.dialog = self.xml.get_widget ("profiles_dialog")
        self.dialog.connect ("destroy", gtk.main_quit)
        self.dialog.set_default_response (gtk.RESPONSE_ACCEPT)

        self.profiles_list = self.xml.get_widget ("profiles_list")
        self.__setup_profiles_list ()
        
        self.profiles_list.connect ("key-press-event", self.__handle_key_press)

        self.new_button = self.xml.get_widget ("new_button")
        self.__fix_button_align (self.new_button)
        self.new_button.connect ("clicked", self.__new_button_clicked)

        self.edit_button = self.xml.get_widget ("edit_button")
        self.__fix_button_align (self.edit_button)
        self.edit_button.connect ("clicked", self.__edit_button_clicked)
        
        self.properties_button = self.xml.get_widget ("properties_button")
        self.__fix_button_align (self.properties_button)
        self.properties_button.connect ("clicked", self.__properties_button_clicked)
        
        self.delete_button = self.xml.get_widget ("delete_button")
        self.__fix_button_align (self.delete_button)
        self.delete_button.connect ("clicked", self.__delete_button_clicked)
        
        self.help_button = self.xml.get_widget ("help_button")
        self.help_button.set_sensitive (False)

        self.dialog.connect ("response", self.__dialog_response)

        (width, height) = self.profiles_list.size_request ()

        self.dialog.set_default_size (min (width + 250, 450),
                                      min (height + 190, 400))

        self.profiles_list.grab_focus ()
        self.__profile_selection_changed (self.profiles_list.get_selection ())

        self.dialog.show ()

    def __fix_button_align (self, button):
        child = button.get_child ()

        if isinstance (child, gtk.Alignment):
            child.set_property ("xalign", 0.0)
        elif isinstance (child, gtk.Label):
            child.set_property ("xalign", 0.0)

    def __setup_profiles_list (self):
        self.profiles_model = ProfilesModel ()
        self.profiles_list.set_model (self.profiles_model)

        self.profiles_list.get_selection ().set_mode (gtk.SELECTION_SINGLE)
        self.profiles_list.get_selection ().connect ("changed", self.__profile_selection_changed)

        c = gtk.TreeViewColumn (_("Name"),
                                gtk.CellRendererText (),
                                text = ProfilesModel.COLUMN_NAME)
        self.profiles_list.append_column (c)

    def __dialog_response (self, dialog, response_id):
        dialog.destroy ()

    def __new_button_clicked (self, button):
        (profile_name, base_profile) = NewProfileDialog (self.profiles_model).run (self.dialog)
        if profile_name:
            self.__create_new_profile (profile_name, base_profile)

    def __get_selected_profile (self):
        (model, row) = self.profiles_list.get_selection ().get_selected ()
        if not row:
            return None
        return model[row][ProfilesModel.COLUMN_NAME]

    def __copy_to_user (self, profile_path, username):
        (fd, user_path) = tempfile.mkstemp (prefix = "profile-%s-" % username, suffix = ".zip")
        os.close (fd)

        shutil.copy2 (profile_path, user_path)

        try:
            pw = pwd.getpwnam (username)
        except KeyError, e:
            errordialog = gtk.MessageDialog (None,
                                             gtk.DIALOG_DESTROY_WITH_PARENT,
                                             gtk.MESSAGE_ERROR,
                                             gtk.BUTTONS_CLOSE,
                                             _("User account '%s' was not found") % username)
            errordialog.format_secondary_text (_("Sabayon requires a special user account '%s' to be present "
                                                 "on this computer. Try again after creating the account (using, "
                                                 "for example, the 'adduser' command)") % username)
                                               
            errordialog.run ()
            errordialog.destroy ()
            raise e
        
        os.chown (user_path, pw.pw_uid, pw.pw_gid)

        dprint ("Copied %s temporarily to %s" % (profile_path, user_path))

        return user_path

    def __copy_from_user (self, user_path, profile_path):
        os.chown (user_path, os.geteuid (), os.getegid ())
        os.rename (user_path, profile_path)
        dprint ("Moved %s back from %s" % (user_path, profile_path))

    def __edit_button_clicked (self, button):
        profile_name = self.__get_selected_profile ()
        if profile_name:
            profile_path = _get_profile_path_for_name (profile_name)

            user_path = self.__copy_to_user (profile_path, PROTOTYPE_USER)

            self.dialog.set_sensitive (False)
            
            argv = SESSION_TOOL_ARGV + [ PROTOTYPE_USER, user_path ]
            dprint ("Running session: %s" % argv)
            os.spawnv (os.P_WAIT, argv[0], argv)
            
            self.dialog.set_sensitive (True)

            self.__copy_from_user (user_path, profile_path)

    def __properties_button_clicked (self, button):
        profile_name = self.__get_selected_profile ()
        if profile_name:
            editorwindow.ProfileEditorWindow (profile_name, self.dialog)

    def __delete_currently_selected (self):
        profile_name = self.__get_selected_profile ()
        if profile_name:
            dprint ("Deleting '%s'", profile_name)
            os.remove (_get_profile_path_for_name (profile_name))
            self.profiles_model.reload ()

    def __delete_button_clicked (self, button):
        self.__delete_currently_selected ()

    def __handle_key_press (self, profiles_list, event):
        if event.keyval in (gtk.keysyms.Delete, gtk.keysyms.KP_Delete):
            self.__delete_currently_selected ()

    def __create_new_profile (self, profile_name, base_profile):
        if base_profile:
            base_storage = storage.ProfileStorage (base_profile)
            new_storage = base_storage.copy (profile_name)
        else:
            new_storage = storage.ProfileStorage (profile_name)
            new_storage.save ()

        self.profiles_model.reload ()
        iter = self.profiles_model.get_iter_first ()
        while iter:
            if self.profiles_model[iter][ProfilesModel.COLUMN_NAME] == profile_name:
                self.profiles_list.get_selection ().select_iter (iter)
                return
            iter = self.profiles_model.iter_next (iter)

    def __profile_selection_changed (self, selection):
        profile_name = self.__get_selected_profile ()
        self.edit_button.set_sensitive (profile_name != None)
        self.properties_button.set_sensitive (profile_name != None)
        self.delete_button.set_sensitive (profile_name != None)



#!/usr/bin/env python

#
# Copyright (C) 2005 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
import os
import os.path

PACKAGE    = "sabayon"
VERSION    = "0.17"
DATADIR    = "/usr/share"
SYSCONFDIR = "/etc"
LIBEXECDIR = "/usr/libexec"
SBINDIR    = "/usr/sbin"

GLADEDIR    = os.path.join (DATADIR, "sabayon", "glade")
PROFILESDIR = os.path.join (SYSCONFDIR, "desktop-profiles")

SESSION_TOOL_ARGV = [ os.path.join (LIBEXECDIR, "sabayon-session") ]
MONITOR_TOOL_ARGV = [ os.path.join (LIBEXECDIR, "sabayon-monitor") ]
APPLY_TOOL_ARGV   = [ os.path.join (SBINDIR,    "sabayon-apply") ]

# FIXME: this is too big a hammer
CLOBBER_USER_PROCESSES_ARGV = [ "/usr/bin/pkill", "-u" ]

# No. of seconds to wait after the session has started before
# running the monitor tool
ADMIN_TOOL_TIMEOUT = 3

#
# This is the user which we run the prototype sessions
# as. The account should be created as part of the packages
# post installation process
#
PROTOTYPE_USER = "sabayon"

DEFAULT_PATH  = "/usr/local/bin:/usr/bin:/bin:/usr/X11R6/bin"
DEFAULT_SHELL = "/bin/bash"
NOLOGIN_SHELL = "/sbin/nologin"
SKEL_HOMEDIR  = "/etc/skel"

XNEST_ARGV   = [ "/usr/X11R6/bin/Xnest", "-terminate", "-audit", "0", "-nolisten", "tcp" ]
SESSION_ARGV = [ "/etc/X11/gdm/Xsession", "gnome" ]
USERMOD_ARGV = [ "/usr/sbin/usermod" ]

PANEL_KEY_BASE = "/apps/panel"
# PANEL_KEY_BASE = "/apps/panel/profiles/default"

#
# When starting the prototype session, which environment
# variables should be passed through umodified
#
PASSTHROUGH_ENVIRONMENT = [
    "SABAYON_DEBUG",
    "LANG",
    "LINGUAS",
    "LC_CTYPE",
    "LC_COLLATE",
    "LC_TIME",
    "LC_NUMERIC",
    "LC_MONETARY",
    "LC_MESSAGES"
]

#
# Files a directories which we don't want to see change
# notifications for
#
FILES_TO_IGNORE = [
    ".xsession-errors",
    ".gnome/gnome-vfs/.trash_entry_cache",
    ".gnome2/share/fonts/fonts.dir",
    ".gnome2/share/cursor-fonts/fonts.dir"
]

DIRECTORIES_TO_IGNORE = [
    ".metacity/sessions",
    ".gconf",
    ".gconf.xml.defaults",
    ".gconf.xml.mandatory",
    ".gconfd",
    ".nautilus/metafiles"
]
